"""Test Liquid filtered expression tokenization.

Filtered expression are those found in output statements, `assign` tags
and `echo` tags.
"""

import unittest

from typing import Any
from typing import NamedTuple

from liquid.expressions.filtered import tokenize
from liquid.expressions.conditional import tokenize as tokenize_conditional
from liquid.expressions.conditional import (
    tokenize_with_parens as tokenize_conditional_with_parens,
)

from liquid.token import TOKEN_COLON
from liquid.token import TOKEN_COMMA
from liquid.token import TOKEN_DOT
from liquid.token import TOKEN_DPIPE
from liquid.token import TOKEN_FLOAT
from liquid.token import TOKEN_IDENTIFIER
from liquid.token import TOKEN_IDENTINDEX
from liquid.token import TOKEN_INTEGER
from liquid.token import TOKEN_LBRACKET
from liquid.token import TOKEN_LPAREN
from liquid.token import TOKEN_PIPE
from liquid.token import TOKEN_RANGE
from liquid.token import TOKEN_RBRACKET
from liquid.token import TOKEN_RPAREN
from liquid.token import TOKEN_STRING
from liquid.token import TOKEN_IF
from liquid.token import TOKEN_TRUE
from liquid.token import TOKEN_ELSE
from liquid.token import TOKEN_NOT
from liquid.token import TOKEN_FALSE
from liquid.token import TOKEN_AND
from liquid.token import TOKEN_LT
from liquid.token import Token


class Case(NamedTuple):
    """Table-driven test helper."""

    description: str
    source: str
    expect: Any


class LexFilteredExpressionTestCase(unittest.TestCase):
    """Test cases for the standard filtered expression."""

    test_cases = [
        Case(
            "string literal single quotes",
            "'foobar'",
            [Token(1, TOKEN_STRING, "foobar")],
        ),
        Case(
            "string literal double quotes",
            "'foobar'",
            [Token(1, TOKEN_STRING, "foobar")],
        ),
        Case(
            "integer literal",
            "7",
            [Token(1, TOKEN_INTEGER, "7")],
        ),
        Case(
            "negative integer literal",
            "-7",
            [
                Token(1, TOKEN_INTEGER, "-7"),
            ],
        ),
        Case(
            "float literal",
            "3.14",
            [Token(1, TOKEN_FLOAT, "3.14")],
        ),
        Case(
            "negative float literal",
            "-3.14",
            [
                Token(1, TOKEN_FLOAT, "-3.14"),
            ],
        ),
        Case(
            "lone identifier",
            "collection",
            [Token(1, TOKEN_IDENTIFIER, "collection")],
        ),
        Case(
            "lone identifier with a hyphen",
            "main-collection",
            [Token(1, TOKEN_IDENTIFIER, "main-collection")],
        ),
        Case(
            "chained identifier",
            "collection.products",
            [
                Token(1, TOKEN_IDENTIFIER, "collection"),
                Token(1, TOKEN_DOT, "."),
                Token(1, TOKEN_IDENTIFIER, "products"),
            ],
        ),
        Case(
            "chained identifier by double quoted key",
            'collection["products"]',
            [
                Token(1, TOKEN_IDENTIFIER, "collection"),
                Token(1, TOKEN_IDENTIFIER, "products"),
            ],
        ),
        Case(
            "chained identifier by single quoted key",
            "collection['products']",
            [
                Token(1, TOKEN_IDENTIFIER, "collection"),
                Token(1, TOKEN_IDENTIFIER, "products"),
            ],
        ),
        Case(
            "chained identifier with array index",
            "collection.products[0]",
            [
                Token(1, TOKEN_IDENTIFIER, "collection"),
                Token(1, TOKEN_DOT, "."),
                Token(1, TOKEN_IDENTIFIER, "products"),
                Token(1, TOKEN_IDENTINDEX, "0"),
            ],
        ),
        Case(
            "chained identifier with array index from identifier",
            "collection.products[i]",
            [
                Token(1, TOKEN_IDENTIFIER, "collection"),
                Token(1, TOKEN_DOT, "."),
                Token(1, TOKEN_IDENTIFIER, "products"),
                Token(1, TOKEN_LBRACKET, "["),
                Token(1, TOKEN_IDENTIFIER, "i"),
                Token(1, TOKEN_RBRACKET, "]"),
            ],
        ),
        Case(
            "string literal with filter",
            "'foo' | upcase",
            [
                Token(1, TOKEN_STRING, "foo"),
                Token(1, TOKEN_PIPE, "|"),
                Token(1, TOKEN_IDENTIFIER, "upcase"),
            ],
        ),
        Case(
            "identifier with filter",
            "collection.title | upcase",
            [
                Token(1, TOKEN_IDENTIFIER, "collection"),
                Token(1, TOKEN_DOT, "."),
                Token(1, TOKEN_IDENTIFIER, "title"),
                Token(1, TOKEN_PIPE, "|"),
                Token(1, TOKEN_IDENTIFIER, "upcase"),
            ],
        ),
        Case(
            "integer literal with filter and integer argument",
            "4 | at_least: 5",
            [
                Token(1, TOKEN_INTEGER, "4"),
                Token(1, TOKEN_PIPE, "|"),
                Token(1, TOKEN_IDENTIFIER, "at_least"),
                Token(1, TOKEN_COLON, ":"),
                Token(1, TOKEN_INTEGER, "5"),
            ],
        ),
        Case(
            "float literal with filter and float argument",
            "4.1 | divided_by: 5.2",
            [
                Token(1, TOKEN_FLOAT, "4.1"),
                Token(1, TOKEN_PIPE, "|"),
                Token(1, TOKEN_IDENTIFIER, "divided_by"),
                Token(1, TOKEN_COLON, ":"),
                Token(1, TOKEN_FLOAT, "5.2"),
            ],
        ),
        Case(
            "string literal with filter and string argument",
            "'foo' | append: 'bar'",
            [
                Token(1, TOKEN_STRING, "foo"),
                Token(1, TOKEN_PIPE, "|"),
                Token(1, TOKEN_IDENTIFIER, "append"),
                Token(1, TOKEN_COLON, ":"),
                Token(1, TOKEN_STRING, "bar"),
            ],
        ),
        Case(
            "string literal with filter and identifier argument",
            "'foo' | append: collection.title",
            [
                Token(1, TOKEN_STRING, "foo"),
                Token(1, TOKEN_PIPE, "|"),
                Token(1, TOKEN_IDENTIFIER, "append"),
                Token(1, TOKEN_COLON, ":"),
                Token(1, TOKEN_IDENTIFIER, "collection"),
                Token(1, TOKEN_DOT, "."),
                Token(1, TOKEN_IDENTIFIER, "title"),
            ],
        ),
        Case(
            "string literal with filter and multiple arguments",
            '"Liquid" | slice: 2, 5',
            [
                Token(1, TOKEN_STRING, "Liquid"),
                Token(1, TOKEN_PIPE, "|"),
                Token(1, TOKEN_IDENTIFIER, "slice"),
                Token(1, TOKEN_COLON, ":"),
                Token(1, TOKEN_INTEGER, "2"),
                Token(1, TOKEN_COMMA, ","),
                Token(1, TOKEN_INTEGER, "5"),
            ],
        ),
        Case(
            "string literal with multiple filters",
            '"Liquid" | slice: 2, 5 | upcase',
            [
                Token(1, TOKEN_STRING, "Liquid"),
                Token(1, TOKEN_PIPE, "|"),
                Token(1, TOKEN_IDENTIFIER, "slice"),
                Token(1, TOKEN_COLON, ":"),
                Token(1, TOKEN_INTEGER, "2"),
                Token(1, TOKEN_COMMA, ","),
                Token(1, TOKEN_INTEGER, "5"),
                Token(1, TOKEN_PIPE, "|"),
                Token(1, TOKEN_IDENTIFIER, "upcase"),
            ],
        ),
        Case(
            "inconsistent whitespace",
            ' "Liquid"   |slice: 2,5',
            [
                Token(1, TOKEN_STRING, "Liquid"),
                Token(1, TOKEN_PIPE, "|"),
                Token(1, TOKEN_IDENTIFIER, "slice"),
                Token(1, TOKEN_COLON, ":"),
                Token(1, TOKEN_INTEGER, "2"),
                Token(1, TOKEN_COMMA, ","),
                Token(1, TOKEN_INTEGER, "5"),
            ],
        ),
        Case(
            "range literal",
            "(1..5)",
            [
                Token(1, TOKEN_LPAREN, "("),
                Token(1, TOKEN_INTEGER, "1"),
                Token(1, TOKEN_RANGE, ".."),
                Token(1, TOKEN_INTEGER, "5"),
                Token(1, TOKEN_RPAREN, ")"),
            ],
        ),
        Case(
            "range literal with float literal start",
            "(2.4..5)",
            [
                Token(1, TOKEN_LPAREN, "("),
                Token(1, TOKEN_FLOAT, "2.4"),
                Token(1, TOKEN_RANGE, ".."),
                Token(1, TOKEN_INTEGER, "5"),
                Token(1, TOKEN_RPAREN, ")"),
            ],
        ),
        Case(
            "range literal with identifiers",
            "(a..b)",
            [
                Token(1, TOKEN_LPAREN, "("),
                Token(1, TOKEN_IDENTIFIER, "a"),
                Token(1, TOKEN_RANGE, ".."),
                Token(1, TOKEN_IDENTIFIER, "b"),
                Token(1, TOKEN_RPAREN, ")"),
            ],
        ),
    ]

    def test_standard_lexer(self) -> None:
        """Test that we can tokenize standard filtered expressions."""
        for case in self.test_cases:
            with self.subTest(msg=case.description):
                tokens = list(tokenize(case.source))
                self.assertEqual(len(tokens), len(case.expect))

                for got, want in zip(tokens, case.expect):
                    self.assertEqual(got, want)

    def test_conditional_lexer(self) -> None:
        """Test that the non-standard conditional expression lexer
        is backwards compatible with standard filtered expressions."""
        for case in self.test_cases:
            with self.subTest(msg=case.description):
                tokens = list(tokenize_conditional(case.source))
                self.assertEqual(len(tokens), len(case.expect))

                for got, want in zip(tokens, case.expect):
                    self.assertEqual(got, want)

    def test_conditional_lexer_with_parens(self) -> None:
        """Test that the non-standard conditional expression lexer, including
        logical `not` and parentheses, is backwards compatible with standard
        filtered expressions."""
        for case in self.test_cases:
            with self.subTest(msg=case.description):
                range_count = len([tok for tok in case.expect if tok[1] == TOKEN_RANGE])
                tokens = list(tokenize_conditional_with_parens(case.source))
                self.assertEqual(len(tokens) - range_count, len(case.expect))


class LexConditionalExpressionTestCase(unittest.TestCase):
    """Test cases for the non-standard filtered expressions with inline conditions."""

    test_cases = [
        Case(
            description="simple condition",
            source="'foo' if true",
            expect=[
                Token(1, TOKEN_STRING, "foo"),
                Token(1, TOKEN_IF, "if"),
                Token(1, TOKEN_TRUE, "true"),
            ],
        ),
        Case(
            description="comparison operator",
            source="'foo' if x < y",
            expect=[
                Token(1, TOKEN_STRING, "foo"),
                Token(1, TOKEN_IF, "if"),
                Token(1, TOKEN_IDENTIFIER, "x"),
                Token(1, TOKEN_LT, "<"),
                Token(1, TOKEN_IDENTIFIER, "y"),
            ],
        ),
        Case(
            description="simple condition with alternative",
            source="'foo' if true else 'bar'",
            expect=[
                Token(1, TOKEN_STRING, "foo"),
                Token(1, TOKEN_IF, "if"),
                Token(1, TOKEN_TRUE, "true"),
                Token(1, TOKEN_ELSE, "else"),
                Token(1, TOKEN_STRING, "bar"),
            ],
        ),
        Case(
            description="condition with preceding filter",
            source="'foo' | upcase if true else 'bar'",
            expect=[
                Token(1, TOKEN_STRING, "foo"),
                Token(1, TOKEN_PIPE, "|"),
                Token(1, TOKEN_IDENTIFIER, "upcase"),
                Token(1, TOKEN_IF, "if"),
                Token(1, TOKEN_TRUE, "true"),
                Token(1, TOKEN_ELSE, "else"),
                Token(1, TOKEN_STRING, "bar"),
            ],
        ),
        Case(
            description="condition with alternative filter",
            source="'foo' if true else 'bar' | upcase",
            expect=[
                Token(1, TOKEN_STRING, "foo"),
                Token(1, TOKEN_IF, "if"),
                Token(1, TOKEN_TRUE, "true"),
                Token(1, TOKEN_ELSE, "else"),
                Token(1, TOKEN_STRING, "bar"),
                Token(1, TOKEN_PIPE, "|"),
                Token(1, TOKEN_IDENTIFIER, "upcase"),
            ],
        ),
        Case(
            description="condition with tail filter",
            source="'foo' if true else 'bar' || upcase",
            expect=[
                Token(1, TOKEN_STRING, "foo"),
                Token(1, TOKEN_IF, "if"),
                Token(1, TOKEN_TRUE, "true"),
                Token(1, TOKEN_ELSE, "else"),
                Token(1, TOKEN_STRING, "bar"),
                Token(1, TOKEN_DPIPE, "||"),
                Token(1, TOKEN_IDENTIFIER, "upcase"),
            ],
        ),
        Case(
            description="multi-line condition with tail filter",
            source="'foo'\nif true\nelse 'bar' || upcase",
            expect=[
                Token(1, TOKEN_STRING, "foo"),
                Token(2, TOKEN_IF, "if"),
                Token(2, TOKEN_TRUE, "true"),
                Token(3, TOKEN_ELSE, "else"),
                Token(3, TOKEN_STRING, "bar"),
                Token(3, TOKEN_DPIPE, "||"),
                Token(3, TOKEN_IDENTIFIER, "upcase"),
            ],
        ),
    ]

    def test_conditional_lexer(self) -> None:
        """Test that the non-standard conditional expression lexer
        is backwards compatible with standard filtered expressions."""
        for case in self.test_cases:
            with self.subTest(msg=case.description):
                tokens = list(tokenize_conditional(case.source))
                self.assertEqual(len(tokens), len(case.expect))

                for got, want in zip(tokens, case.expect):
                    self.assertEqual(got, want)

    def test_conditional_lexer_with_parens(self) -> None:
        """Test that the non-standard conditional expression lexer, including
        logical `not` and parentheses, is backwards compatible with standard
        filtered expressions."""
        for case in self.test_cases:
            with self.subTest(msg=case.description):
                range_count = len([tok for tok in case.expect if tok[1] == TOKEN_RANGE])
                tokens = list(tokenize_conditional_with_parens(case.source))
                self.assertEqual(len(tokens) - range_count, len(case.expect))


class LexConditionalNotExpressionTestCase(unittest.TestCase):
    """Test cases for the non-standard filtered expressions with inline conditions
    and logical `not` with parentheses for grouping terms."""

    test_cases = [
        Case(
            description="negated condition",
            source="'foo' if not true",
            expect=[
                Token(1, TOKEN_STRING, "foo"),
                Token(1, TOKEN_IF, "if"),
                Token(1, TOKEN_NOT, "not"),
                Token(1, TOKEN_TRUE, "true"),
            ],
        ),
        Case(
            description="negated condition with alternative",
            source="'foo' if not true else 'bar'",
            expect=[
                Token(1, TOKEN_STRING, "foo"),
                Token(1, TOKEN_IF, "if"),
                Token(1, TOKEN_NOT, "not"),
                Token(1, TOKEN_TRUE, "true"),
                Token(1, TOKEN_ELSE, "else"),
                Token(1, TOKEN_STRING, "bar"),
            ],
        ),
        Case(
            description="grouped condition with alternative",
            source="'foo' if not (false and false) else 'bar'",
            expect=[
                Token(1, TOKEN_STRING, "foo"),
                Token(1, TOKEN_IF, "if"),
                Token(1, TOKEN_NOT, "not"),
                Token(1, TOKEN_LPAREN, "("),
                Token(1, TOKEN_FALSE, "false"),
                Token(1, TOKEN_AND, "and"),
                Token(1, TOKEN_FALSE, "false"),
                Token(1, TOKEN_RPAREN, ")"),
                Token(1, TOKEN_ELSE, "else"),
                Token(1, TOKEN_STRING, "bar"),
            ],
        ),
        Case(
            description="condition with preceding filter",
            source="'foo' | upcase if not true else 'bar'",
            expect=[
                Token(1, TOKEN_STRING, "foo"),
                Token(1, TOKEN_PIPE, "|"),
                Token(1, TOKEN_IDENTIFIER, "upcase"),
                Token(1, TOKEN_IF, "if"),
                Token(1, TOKEN_NOT, "not"),
                Token(1, TOKEN_TRUE, "true"),
                Token(1, TOKEN_ELSE, "else"),
                Token(1, TOKEN_STRING, "bar"),
            ],
        ),
    ]

    def test_conditional_lexer_with_parens(self) -> None:
        """Test that the non-standard conditional expression lexer, including
        logical `not` and parentheses, is backwards compatible with standard
        filtered expressions."""
        for case in self.test_cases:
            with self.subTest(msg=case.description):
                tokens = list(tokenize_conditional_with_parens(case.source))
                self.assertEqual(len(tokens), len(case.expect))

                for got, want in zip(tokens, case.expect):
                    self.assertEqual(got, want)



"""Golden tests cases for testing liquid's built-in `reverse` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="array of strings",
        template=r"{{ a | reverse | join: '#' }}",
        expect="A#B#a#b",
        globals={"a": ["b", "a", "B", "A"]},
    ),
    Case(
        description="array of things",
        template=r"{{ a | reverse | join: '#' }}",
        expect=r"{}#1#b#a",
        globals={"a": ["a", "b", 1, [], {}]},
    ),
    Case(
        description="empty array",
        template=r"{{ a | reverse | join: '#' }}",
        expect="",
        globals={"a": []},
    ),
    Case(
        description="unexpected argument",
        template=r"{{ a | reverse: 0 | join: '#' }}",
        expect="",
        globals={"a": []},
        error=True,
    ),
    Case(
        description="left value not an array",
        template=r"{{ a | reverse | join: '#' }}",
        expect="123",
        globals={"a": 123},
    ),
    Case(
        description="left value is undefined",
        template=r"{{ nosuchthing | reverse | join: '#' }}",
        expect="",
    ),
]



"""Tag and output statement expression objects."""

from __future__ import annotations

import sys

from abc import ABC
from abc import abstractmethod

from collections import abc
from decimal import Decimal
from itertools import islice

from typing import Any
from typing import Dict
from typing import Generic
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Mapping
from typing import Optional
from typing import Tuple
from typing import TypeVar
from typing import Union

try:
    from markupsafe import Markup
except ImportError:
    from liquid.exceptions import Markup  # type: ignore

from liquid.context import Context

from liquid.exceptions import LiquidTypeError
from liquid.exceptions import Error
from liquid.exceptions import NoSuchFilterFunc
from liquid.exceptions import FilterValueError

from liquid.limits import to_int

# pylint: disable=missing-class-docstring too-few-public-methods too-many-lines


class Expression(ABC):
    __slots__ = ()

    @abstractmethod
    def evaluate(self, context: Context) -> object:
        """Evaluate the expression with the given context."""

    async def evaluate_async(self, context: Context) -> object:
        """An async version of :meth:`liquid.expression.Expression.evaluate`."""
        return self.evaluate(context)

    def children(self) -> List[Expression]:
        """Return a list of child expressions."""
        raise NotImplementedError(f"{self.__class__.__name__}.children")


class Nil(Expression):
    __slots__ = ()

    def __eq__(self, other: object) -> bool:
        return other is None or isinstance(other, Nil)

    def __repr__(self) -> str:  # pragma: no cover
        return "NIL()"

    def __str__(self) -> str:  # pragma: no cover
        return ""

    def evaluate(self, context: Context) -> None:
        return None

    def children(self) -> List[Expression]:
        return []


NIL = Nil()


class Empty(Expression):
    __slots__ = ()

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Empty):
            return True
        if isinstance(other, (list, dict, str)) and not other:
            return True
        return False

    def __repr__(self) -> str:  # pragma: no cover
        return "Empty()"

    def __str__(self) -> str:  # pragma: no cover
        return "empty"

    def evaluate(self, context: Context) -> Empty:
        return self

    def children(self) -> List[Expression]:
        return []


EMPTY = Empty()


class Blank(Expression):
    __slots__ = ()

    def __eq__(self, other: object) -> bool:
        if isinstance(other, str) and (not other or other.isspace()):
            return True
        if isinstance(other, (list, dict)) and not other:
            return True
        if isinstance(other, Blank):
            return True
        return False

    def __repr__(self) -> str:  # pragma: no cover
        return "Blank()"

    def __str__(self) -> str:  # pragma: no cover
        return "blank"

    def evaluate(self, context: Context) -> Blank:
        return self

    def children(self) -> List[Expression]:
        return []


BLANK = Blank()


class Continue(Expression):
    __slots__ = ()

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Continue):
            return True
        return False

    def __repr__(self) -> str:  # pragma: no cover
        return "Continue()"

    def __str__(self) -> str:  # pragma: no cover
        return "continue"

    def evaluate(self, context: Context) -> int:
        return 0

    def children(self) -> List[Expression]:
        return []


CONTINUE = Continue()


T = TypeVar("T")  # pylint: disable=invalid-name


class Literal(Expression, Generic[T]):
    __slots__ = ("value",)

    def __init__(self, value: T):
        self.value = value

    def __str__(self) -> str:
        return repr(self.value)

    def __eq__(self, other: object) -> bool:
        return self.value == other

    def __hash__(self) -> int:
        return hash(self.value)

    def __sizeof__(self) -> int:
        return sys.getsizeof(self.value)

    def evaluate(self, context: Context) -> object:
        return self.value

    def children(self) -> List[Expression]:
        return []


class Boolean(Literal[bool]):
    __slots__ = ()

    def __init__(self, value: bool):
        super().__init__(value)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Boolean) and self.value == other.value

    def __repr__(self) -> str:  # pragma: no cover
        return f"Boolean(value={self.value})"


TRUE = Boolean(True)
FALSE = Boolean(False)


class StringLiteral(Literal[str]):
    __slots__ = ()

    def __init__(self, value: str):
        super().__init__(value)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, StringLiteral) and self.value == other.value

    def __hash__(self) -> int:
        return hash(self.value)

    def __repr__(self) -> str:  # pragma: no cover
        return f"StringLiteral(value='{self.value}')"

    def __sizeof__(self) -> int:
        return sys.getsizeof(self.value)

    def evaluate(self, context: Context) -> Union[str, Markup]:
        if context.autoescape:
            return Markup(self.value)
        return self.value


class IntegerLiteral(Literal[int]):
    __slots__ = ()

    def __init__(self, value: int):
        super().__init__(value)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, IntegerLiteral) and self.value == other.value

    def __hash__(self) -> int:
        return hash(self.value)

    def __repr__(self) -> str:  # pragma: no cover
        return f"IntegerLiteral(value={self.value})"


class FloatLiteral(Literal[float]):
    __slots__ = ()

    def __init__(self, value: float):
        super().__init__(value)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, FloatLiteral) and self.value == other.value

    def __repr__(self) -> str:  # pragma: no cover
        return f"FloatLiteral(value={self.value})"


class RangeLiteral(Expression):
    __slots__ = ("start", "stop")

    def __init__(self, start: Expression, stop: Expression):
        self.start = start
        self.stop = stop

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, RangeLiteral)
            and self.start == other.start
            and self.stop == other.stop
        )

    def __str__(self) -> str:
        return f"({self.start}..{self.stop})"

    def __repr__(self) -> str:  # pragma: no cover
        return f"RangeLiteral(start={self.start}, stop={self.stop})"

    def __hash__(self) -> int:
        return hash((self.start, self.stop))

    def __sizeof__(self) -> int:
        return (
            super().__sizeof__() + sys.getsizeof(self.start) + sys.getsizeof(self.stop)
        )

    def _make_range(self, start: Any, stop: Any) -> range:
        try:
            start = to_int(start)
        except ValueError:
            start = 0

        try:
            stop = to_int(stop)
        except ValueError:
            stop = 0

        # Descending ranges don't work
        if start > stop:
            return range(0)

        return range(start, stop + 1)

    def evaluate(self, context: Context) -> range:
        return self._make_range(
            self.start.evaluate(context), self.stop.evaluate(context)
        )

    async def evaluate_async(self, context: Context) -> range:
        return self._make_range(
            await self.start.evaluate_async(context),
            await self.stop.evaluate_async(context),
        )

    def children(self) -> List[Expression]:
        return [self.start, self.stop]


class IdentifierPathElement(Literal[Union[int, str]]):
    __slots__ = ()

    def __eq__(self, other: object) -> bool:
        return isinstance(other, IdentifierPathElement) and self.value == other.value

    def __hash__(self) -> int:
        return hash(self.value)

    def __repr__(self) -> str:  # pragma: no cover
        return f"IdentifierPathElement(value={self.value})"

    def __str__(self) -> str:
        return str(self.value)


IdentifierPath = List[Union[IdentifierPathElement, "Identifier"]]


class Identifier(Expression):
    __slots__ = ("path",)

    def __init__(self, path: IdentifierPath):
        self.path = path

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Identifier) and self.path == other.path

    def __repr__(self) -> str:  # pragma: no cover
        return f"Identifier(path={self.path})"

    def __str__(self) -> str:
        buf = []
        for elem in self.path:
            if isinstance(elem, Identifier):
                buf.append(f"[{elem}]")
            else:
                buf.append(str(elem))
        return ".".join(buf)

    def __hash__(self) -> int:
        return hash(tuple(self.path))

    def __sizeof__(self) -> int:
        return super().__sizeof__() + sys.getsizeof(self.path)

    def evaluate(self, context: Context) -> object:
        path: List[Any] = [elem.evaluate(context) for elem in self.path]
        return context.get(path)

    async def evaluate_async(self, context: Context) -> object:
        path: List[Any] = [await elem.evaluate_async(context) for elem in self.path]
        return await context.get_async(path)

    def children(self) -> List[Expression]:
        return [el for el in self.path if isinstance(el, Expression)]


class PrefixExpression(Expression):
    __slots__ = ("operator", "right")

    def __init__(self, operator: str, right: Expression):
        self.operator = operator
        self.right = right

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, PrefixExpression)
            and self.operator == other.operator
            and self.right == other.right
        )

    def __repr__(self) -> str:  # pragma: no cover
        return f"PrefixExpression(operator='{self.operator}', right={self.right!r})"

    def __str__(self) -> str:
        return f"({self.operator}{self.right})"

    def _evaluate(self, right: object) -> Union[int, float]:
        if self.operator == "not":
            return not is_truthy(right)
        if self.operator == "-":
            if isinstance(right, (int, float)):
                return -right
            raise LiquidTypeError(f"unknown operator {self.operator}{self.right}")

        raise LiquidTypeError(f"unknown operator {self.operator}")

    def evaluate(self, context: Context) -> object:
        return self._evaluate(self.right.evaluate(context))

    async def evaluate_async(self, context: Context) -> object:
        return self._evaluate(await self.right.evaluate_async(context))

    def children(self) -> List[Expression]:
        return [self.right]


class InfixExpression(Expression):
    __slots__ = ("left", "operator", "right")

    def __init__(
        self,
        left: Expression,
        operator: str,
        right: Expression,
    ):
        self.left = left
        self.operator = operator
        self.right = right

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, InfixExpression)
            and self.left == other.left
            and self.operator == other.operator
            and self.right == other.right
        )

    def __repr__(self) -> str:  # pragma: no cover
        return (
            f"InfixExpression(left={self.left!r}, "
            f"operator='{self.operator}', right={self.right!r})"
        )

    def __str__(self) -> str:
        return f"({self.left} {self.operator} {self.right})"

    def evaluate(self, context: Context) -> object:
        return compare(
            self.left.evaluate(context), self.operator, self.right.evaluate(context)
        )

    async def evaluate_async(self, context: Context) -> object:
        return compare(
            await self.left.evaluate_async(context),
            self.operator,
            await self.right.evaluate_async(context),
        )

    def children(self) -> List[Expression]:
        return [self.left, self.right]


class Filter:
    __slots__ = ("name", "args", "kwargs")

    def __init__(
        self,
        name: str,
        args: List[Expression],
        kwargs: Optional[Mapping[str, Expression]] = None,
    ):
        self.name = name
        self.args = args
        self.kwargs = kwargs or {}

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, Filter)
            and self.name == other.name
            and self.args == other.args
            and self.kwargs == other.kwargs
        )

    def __repr__(self) -> str:  # pragma: no cover
        return f"Filter(name='{self.name}', args={self.args}, kwargs={self.kwargs})"

    def __str__(self) -> str:
        buf = [self.name]

        args_str = ", ".join([str(arg) for arg in self.args])
        if args_str:
            buf.append(f": {args_str}")
            # return f"{self.name}: {args_str})"

        kwargs_str = ", ".join([f"{k}: {v}" for k, v in self.kwargs.items()])
        if kwargs_str:
            buf.append(f", {kwargs_str}")

        return "".join(buf)

    def evaluate_args(self, context: Context) -> List[object]:
        """Return a list of filter arguments evaluated with the given context."""
        return [arg.evaluate(context) for arg in self.args]

    async def evaluate_args_async(self, context: Context) -> List[object]:
        """An async version of `evaluate_args`."""
        return [await arg.evaluate_async(context) for arg in self.args]

    def evaluate_kwargs(self, context: Context) -> Dict[str, object]:
        """Return a dictionary of filter keyword arguments evaluated with the given
        context."""
        # Shortcut for the common case. Most filters do not use named parameters.
        if not self.kwargs:
            return {}
        return {k: v.evaluate(context) for k, v in self.kwargs.items()}

    async def evaluate_kwargs_async(self, context: Context) -> Dict[str, object]:
        """An async version of `evaluate_kwargs`."""
        # Shortcut for the common case. Most filters do not use named parameters.
        if not self.kwargs:
            return {}
        return {k: await v.evaluate_async(context) for k, v in self.kwargs.items()}


class BooleanExpression(Expression):
    __slots__ = ("expression",)

    def __init__(self, expression: Expression):
        self.expression = expression

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, BooleanExpression) and self.expression == other.expression
        )

    def __str__(self) -> str:
        expr_str = str(self.expression)
        if not expr_str.startswith("("):
            return f"({expr_str})"
        return expr_str

    def __repr__(self) -> str:  # pragma: no cover
        return f"BooleanExpression(expression={self.expression!r})"

    def evaluate(self, context: Context) -> bool:
        return is_truthy(self.expression.evaluate(context))

    async def evaluate_async(self, context: Context) -> bool:
        return is_truthy(await self.expression.evaluate_async(context))

    def children(self) -> List[Expression]:
        return [self.expression]


class FilteredExpression(Expression):
    __slots__ = ("expression", "filters")

    def __init__(self, expression: Expression, filters: Optional[List[Filter]] = None):
        self.expression = expression
        self.filters = filters or []

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, FilteredExpression)
            and self.expression == other.expression
            and self.filters == other.filters
        )

    def __str__(self) -> str:
        filter_str = " | ".join([str(filter) for filter in self.filters])

        if filter_str:
            return f"{self.expression} | {filter_str}"
        return str(self.expression)

    def __repr__(self) -> str:  # pragma: no cover
        return (
            f"FilteredExpression(expression={self.expression!r}, "
            f"filters={self.filters})"
        )

    def apply_filters(
        self,
        left: object,
        filters: Iterable[Filter],
        context: Context,
    ) -> object:
        """Return the result of applying all filters in this expression to an object."""
        rv = left
        for _filter in filters:
            try:
                func = context.filter(_filter.name)
            except NoSuchFilterFunc:
                if context.env.strict_filters:
                    raise
                continue

            # Any exception causes us to abort the filter chain and discard the result.
            # Nothing will be rendered.
            try:
                if not _filter.args and not _filter.kwargs:
                    rv = func(rv)
                elif _filter.args and not _filter.kwargs:
                    rv = func(rv, *_filter.evaluate_args(context))
                else:
                    rv = func(
                        rv,
                        *_filter.evaluate_args(context),
                        **_filter.evaluate_kwargs(context),
                    )
            except FilterValueError:
                # Pass over filtered expressions who's left value is not allowed.
                continue
            except Error:
                raise
            except Exception as err:
                raise Error(
                    f"filter '{_filter.name}': unexpected error: {err}"
                ) from err

        return rv

    async def apply_filters_async(
        self,
        left: object,
        filters: Iterable[Filter],
        context: Context,
    ) -> object:
        """Return the result of applying all filters in this expression to an object."""
        rv = left
        for _filter in filters:
            try:
                func = context.filter(_filter.name)
            except NoSuchFilterFunc:
                if context.env.strict_filters:
                    raise
                continue

            # Any exception causes us to abort the filter chain and discard the result.
            # Nothing will be rendered.
            try:
                if not _filter.args and not _filter.kwargs:
                    rv = func(rv)
                elif _filter.args and not _filter.kwargs:
                    rv = func(
                        rv,
                        *(await _filter.evaluate_args_async(context)),
                    )
                else:
                    rv = func(
                        rv,
                        *(await _filter.evaluate_args_async(context)),
                        **(await _filter.evaluate_kwargs_async(context)),
                    )
            except FilterValueError:
                # Pass over filtered expressions who's left value is not allowed.
                continue
            except Error:
                raise
            except Exception as err:
                raise Error(
                    f"filter '{_filter.name}': unexpected error: {err}"
                ) from err

        return rv

    def evaluate(self, context: Context) -> object:
        return self.apply_filters(
            self.expression.evaluate(context),
            self.filters,
            context,
        )

    async def evaluate_async(self, context: Context) -> object:
        return await self.apply_filters_async(
            await self.expression.evaluate_async(context),
            self.filters,
            context,
        )

    def children(self) -> List[Expression]:
        _children = [self.expression]
        for _filter in self.filters:
            _children.extend(_filter.args)
            _children.extend(_filter.kwargs.values())
        return _children


class ConditionalExpression(FilteredExpression):
    __slots__ = (
        "condition",
        "alternative",
    )

    def __init__(
        self,
        expression: Expression,
        filters: Optional[List[Filter]] = None,
        condition: Optional[Expression] = None,
        alternative: Optional[Expression] = None,
    ):
        super().__init__(expression, filters)
        self.expression = expression
        self.condition = condition
        self.alternative = alternative
        self.filters = filters or []  # These are "tail" filters

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, ConditionalExpression)
            and self.expression == other.expression
            and self.condition == other.condition
            and self.alternative == other.alternative
            and self.filters == other.filters
        )

    def __str__(self) -> str:
        buf = [str(self.expression)]

        if self.condition:
            buf.append("if")
            buf.append(str(self.condition))

        if self.alternative:
            buf.append("else")
            buf.append(str(self.alternative))

        if self.filters:
            buf.append("|")
            buf.append(" | ".join([str(filter) for filter in self.filters]))

        return " ".join(buf)

    def __repr__(self) -> str:  # pragma: no cover
        return (
            f"ConditionalExpression(expression={self.expression!r}, "
            f"condition={self.condition}, "
            f"alternative={self.alternative}, "
            f"filters={self.filters})"
        )

    def evaluate(self, context: Context) -> object:
        if self.condition:
            if is_truthy(self.condition.evaluate(context)):
                result = self.expression.evaluate(context)

            elif self.alternative:
                result = self.alternative.evaluate(context)
            else:
                result = context.env.undefined("")
        else:
            result = self.expression.evaluate(context)
            if self.filters:
                result = self.apply_filters(result, self.filters, context)

        if self.filters:
            result = self.apply_filters(result, self.filters, context)

        return result

    async def evaluate_async(self, context: Context) -> object:
        if self.condition:
            if is_truthy(await self.condition.evaluate_async(context)):
                result = await self.expression.evaluate_async(context)

            elif self.alternative:
                result = await self.alternative.evaluate_async(context)
            else:
                result = context.env.undefined("")
        else:
            result = await self.expression.evaluate_async(context)
            if self.filters:
                result = await self.apply_filters_async(result, self.filters, context)

        if self.filters:
            result = await self.apply_filters_async(result, self.filters, context)

        return result

    def children(self) -> List[Expression]:
        _children = [self.expression]

        if self.condition is not None:
            _children.append(self.condition)

        if self.alternative is not None:
            _children.append(self.alternative)

        for _filter in self.filters:
            _children.extend(_filter.args)
            _children.extend(_filter.kwargs.values())

        return _children


class AssignmentExpression(Expression):
    __slots__ = ("name", "expression")

    def __init__(self, name: str, expression: Expression):
        self.name = name
        self.expression = expression

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, AssignmentExpression)
            and self.name == other.name
            and self.expression == other.expression
        )

    def __str__(self) -> str:
        return f"{self.name} = {self.expression}"

    def __repr__(self) -> str:  # pragma: no cover
        return (
            f"AssignmentExpression(name='{self.name}', expression={self.expression!r})"
        )

    def evaluate(self, context: Context) -> str:
        context.assign(key=self.name, val=self.expression.evaluate(context))
        return ""

    async def evaluate_async(self, context: Context) -> str:
        context.assign(
            key=self.name,
            val=(await self.expression.evaluate_async(context)),
        )
        return ""

    def children(self) -> List[Expression]:
        return [self.expression]


LoopArgument = Union[
    IntegerLiteral,
    FloatLiteral,
    Identifier,
    Continue,
    Nil,
]

# An identifier that resolves to an iterable or a range expression.
LoopIterable = Union[Identifier, RangeLiteral]


class LoopExpression(Expression):
    """Loop expression tree/object model, as returned by parse_loop_expression.

    Start, stop, limit and offset can be integer literals or identifiers
    that resolve to an integer at render time.

    `cols` is used by the built-in tag `tablerow` only. It will be ignored
    by the `for` tag.
    """

    __slots__ = (
        "name",
        "iterable",
        "limit",
        "offset",
        "cols",
        "reversed",
    )

    def __init__(
        self,
        name: str,
        iterable: LoopIterable,
        limit: LoopArgument = NIL,
        offset: LoopArgument = NIL,
        cols: LoopArgument = NIL,
        reversed_: bool = False,
    ):
        self.name = name
        self.iterable = iterable
        self.limit = limit
        self.offset = offset
        self.cols = cols
        self.reversed = reversed_

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, LoopExpression)
            and self.name == other.name
            and self.iterable == other.iterable
            and self.limit == other.limit
            and self.offset == other.offset
            and self.cols == other.cols
            and self.reversed == other.reversed
        )

    def __str__(self) -> str:
        buf = [f"{self.name} in", str(self.iterable)]

        if self.limit != NIL:
            buf.append(f"limit:{self.limit}")

        if self.offset != NIL:
            buf.append(f"offset:{self.offset}")

        if self.cols != NIL:
            buf.append(f"cols:{self.cols}")

        if self.reversed:
            buf.append("reversed")

        return " ".join(buf)

    def __repr__(self) -> str:  # pragma: no cover
        return (
            f"LoopExpression(name='{self.name}', iterable={self.iterable}, "
            f"limit={self.limit}, offset={self.offset}, "
            f"cols={self.cols}, reversed={self.reversed})"
        )

    def _limit_and_offset(
        self,
        it: Iterator[Any],
        size: int,
        limit: Optional[int],
        offset: Optional[int | Continue],
        context: Context,
    ) -> Tuple[Iterator[Any], int]:
        """Slice iterable `it` according to `limit` and `offset`."""
        length = size

        # For the sake of the special for loop offset `continue`, we need to derive an
        # identifier for this loop and store the theoretical stop index on the render
        # context using that identifier.
        #
        # For compatibility with the reference implementation, we'll key loop stop
        # indexes on the loop item name and the name of the iterable. Not the contents
        # of the iterable. In the case of a range expression, we'll use a string
        # representation of the expression.
        offset_key = "-".join([self.name, str(self.iterable)])

        if limit is None and offset is None:
            context.stopindex(key=offset_key, index=length)
            if self.reversed:
                return reversed(list(it)), length
            return it, length

        if offset == CONTINUE:
            offset = context.stopindex(key=offset_key)
            length = max(length - offset, 0)
        elif offset is not None:
            assert isinstance(offset, int), f"found {offset!r}"
            length = max(length - offset, 0)

        if limit is not None:
            length = min(length, limit)

        if offset:
            stop = offset + length
        else:
            stop = length

        context.stopindex(key=offset_key, index=stop)
        it = islice(it, offset, stop)

        if self.reversed:
            return reversed(list(it)), length
        return it, length

    def _obj_to_iter(self, obj: object) -> Tuple[Iterator[Any], int]:
        if isinstance(obj, abc.Mapping):
            return iter(obj.items()), len(obj)
        if isinstance(obj, range):
            return iter(obj), len(obj)
        if isinstance(obj, abc.Sequence):
            return iter(obj), len(obj)

        raise LiquidTypeError(
            f"expected array or hash at '{self.iterable}', found '{str(obj)}'"
        )

    def _raise_for_invalid_argument(self, arg: object) -> None:
        if arg is not None and not isinstance(arg, int):
            raise LiquidTypeError(f"expected an integer argument, found {arg!r}")

    def evaluate(self, context: Context) -> Tuple[Iterator[Any], int]:
        obj = self.iterable.evaluate(context)
        it, length = self._obj_to_iter(obj)

        limit = self.limit.evaluate(context)
        self._raise_for_invalid_argument(limit)
        assert isinstance(limit, int) or limit is None

        if self.offset != CONTINUE:
            offset = self.offset.evaluate(context)
            self._raise_for_invalid_argument(offset)
            assert isinstance(offset, int) or offset is None
        else:
            offset = self.offset
            assert isinstance(offset, Continue)

        return self._limit_and_offset(
            it,
            size=length,
            limit=limit,
            offset=offset,
            context=context,
        )

    async def evaluate_async(self, context: Context) -> Tuple[Iterator[Any], int]:
        obj = await self.iterable.evaluate_async(context)
        it, length = self._obj_to_iter(obj)

        limit = await self.limit.evaluate_async(context)
        self._raise_for_invalid_argument(limit)
        assert isinstance(limit, int) or limit is None

        if self.offset != CONTINUE:
            offset = await self.offset.evaluate_async(context)
            self._raise_for_invalid_argument(offset)
            assert isinstance(offset, int) or offset is None
        else:
            offset = self.offset
            assert isinstance(offset, Continue)

        return self._limit_and_offset(
            it,
            size=length,
            limit=limit,
            offset=offset,
            context=context,
        )

    def children(self) -> List[Expression]:
        return [
            self.iterable,
            self.limit,
            self.offset,
            self.cols,
        ]


Number = Union[int, float]


def eval_number_expression(left: Number, operator: str, right: Number) -> bool:
    """Return the result of comparing two numbers with the given operator."""
    if operator == "<=":
        return left <= right
    if operator == ">=":
        return left >= right
    if operator == "<":
        return left < right
    if operator == ">":
        return left > right

    raise LiquidTypeError(f"unknown operator {left} {operator} {right}")


def _is_py_falsy_number(obj: object) -> bool:
    # Liquid 0, 0.0, 0b0, 0X0, 0o0 and Decimal("0") are not falsy.
    return not isinstance(obj, bool) and isinstance(obj, (int, float, Decimal))


def is_truthy(obj: Any) -> bool:
    """Return True if the given object is Liquid truthy."""
    if hasattr(obj, "__liquid__"):
        obj = obj.__liquid__()
    return _is_py_falsy_number(obj) or obj not in (False, None)


# pylint: disable=too-many-return-statements
def compare_bool(left: Any, operator: str, right: Any) -> bool:
    """Compare an object to a boolean value."""
    if (isinstance(left, bool) and _is_py_falsy_number(right)) or (
        isinstance(right, bool) and _is_py_falsy_number(left)
    ):
        if operator in ("==", "<", ">", "<=", ">="):
            return False
        if operator in ("!=", "<>"):
            return True
        raise LiquidTypeError(
            f"unknown operator: {type(left)} {operator} {type(right)}"
        )

    if operator == "==":
        return bool(left == right)
    if operator in ("!=", "<>"):
        return bool(left != right)
    if operator in ("<", ">", "<=", ">="):
        return False

    raise LiquidTypeError(f"unknown operator: {type(left)} {operator} {type(right)}")


# pylint: disable=too-many-return-statements,too-many-branches
def compare(left: Any, operator: str, right: Any) -> bool:
    """Return the result of a comparison operation between two objects."""
    if operator == "and":
        return is_truthy(left) and is_truthy(right)
    if operator == "or":
        return is_truthy(left) or is_truthy(right)

    if hasattr(left, "__liquid__"):
        left = left.__liquid__()
    if hasattr(right, "__liquid__"):
        right = right.__liquid__()

    if isinstance(right, (Empty, Blank)):
        left, right = right, left

    if isinstance(left, bool) or isinstance(right, bool):
        return compare_bool(left, operator, right)

    if operator == "==":
        return bool(left == right)
    if operator in ("!=", "<>"):
        return bool(left != right)

    if operator == "contains":
        if isinstance(left, str):
            return str(right) in left
        if isinstance(left, (list, dict)):
            return right in left

    if None in (left, right):
        return False

    if type(left) in (int, float) and type(right) in (int, float):
        return eval_number_expression(left, operator, right)

    # pylint: disable=unidiomatic-typecheck
    if type(left) != type(right):
        raise LiquidTypeError(
            f"invalid operator for types '{str(left)} {operator} {str(right)}'"
        )

    raise LiquidTypeError(f"unknown operator: {type(left)} {operator} {type(right)}")



"""Extra HTML filters."""
import html

from liquid import Markup
from liquid.filter import string_filter
from liquid.filter import with_environment
from liquid import Environment


@string_filter
@with_environment
def stylesheet_tag(url: str, *, environment: Environment) -> str:
    """Wrap a URL in an HTML stylesheet tag."""
    tag = '<link href="{}" rel="stylesheet" type="text/css" media="all" />'
    if environment.autoescape:
        return Markup(tag).format(str(url))
    return tag.format(html.escape(url))


@string_filter
@with_environment
def script_tag(url: str, *, environment: Environment) -> str:
    """Wrap a URL in an HTML script tag."""
    tag = '<script src="{}" type="text/javascript"></script>'
    if environment.autoescape:
        return Markup(tag).format(str(url))
    return tag.format(html.escape(url))



"""Functions for parsing boolean expressions. Like those found in `if` and
unless` tags.
"""
from typing import Dict
from typing import Callable

from liquid.expression import Expression
from liquid.expression import BooleanExpression
from liquid.expression import InfixExpression
from liquid.expression import PrefixExpression

from liquid.expressions.common import parse_blank
from liquid.expressions.common import parse_boolean
from liquid.expressions.common import parse_empty
from liquid.expressions.common import parse_float_literal
from liquid.expressions.common import parse_identifier
from liquid.expressions.common import parse_integer_literal
from liquid.expressions.common import parse_nil
from liquid.expressions.common import parse_string_literal
from liquid.expressions.common import make_parse_range

from liquid.expressions.filtered.parse import parse_obj as parse_simple_obj
from liquid.expressions.boolean.lex import tokenize
from liquid.expressions.boolean.lex import tokenize_with_parens
from liquid.expressions.stream import TokenStream

from liquid.exceptions import LiquidSyntaxError

from liquid.token import TOKEN_EOF
from liquid.token import TOKEN_FALSE
from liquid.token import TOKEN_TRUE
from liquid.token import TOKEN_NIL
from liquid.token import TOKEN_NULL
from liquid.token import TOKEN_EMPTY
from liquid.token import TOKEN_BLANK
from liquid.token import TOKEN_STRING
from liquid.token import TOKEN_INTEGER
from liquid.token import TOKEN_FLOAT
from liquid.token import TOKEN_IDENTIFIER
from liquid.token import TOKEN_LPAREN
from liquid.token import TOKEN_RPAREN
from liquid.token import TOKEN_RANGE_LITERAL

from liquid.token import TOKEN_EQ
from liquid.token import TOKEN_OR
from liquid.token import TOKEN_AND
from liquid.token import TOKEN_NOT
from liquid.token import TOKEN_LT
from liquid.token import TOKEN_GT
from liquid.token import TOKEN_NE
from liquid.token import TOKEN_LG
from liquid.token import TOKEN_LE
from liquid.token import TOKEN_GE
from liquid.token import TOKEN_CONTAINS

# Note that PREFIX and LOGICAL are not used in any "standard" expressions.
PRECEDENCE_LOWEST = 1
PRECEDENCE_LOGICALRIGHT = 3
PRECEDENCE_LOGICAL = 4
PRECEDENCE_RELATIONAL = 5
PRECEDENCE_MEMBERSHIP = 6
PRECEDENCE_PREFIX = 7

PRECEDENCES = {
    TOKEN_IDENTIFIER: PRECEDENCE_LOWEST,
    TOKEN_EQ: PRECEDENCE_RELATIONAL,
    TOKEN_LT: PRECEDENCE_RELATIONAL,
    TOKEN_GT: PRECEDENCE_RELATIONAL,
    TOKEN_NE: PRECEDENCE_RELATIONAL,
    TOKEN_LG: PRECEDENCE_RELATIONAL,
    TOKEN_LE: PRECEDENCE_RELATIONAL,
    TOKEN_GE: PRECEDENCE_RELATIONAL,
    TOKEN_CONTAINS: PRECEDENCE_MEMBERSHIP,
    TOKEN_AND: PRECEDENCE_LOGICALRIGHT,
    TOKEN_OR: PRECEDENCE_LOGICALRIGHT,
    TOKEN_NOT: PRECEDENCE_LOGICALRIGHT,
    TOKEN_RPAREN: PRECEDENCE_LOWEST,
}


TOKEN_MAP: Dict[str, Callable[[TokenStream], Expression]] = {
    TOKEN_FALSE: parse_boolean,
    TOKEN_TRUE: parse_boolean,
    TOKEN_NIL: parse_nil,
    TOKEN_NULL: parse_nil,
    TOKEN_EMPTY: parse_empty,
    TOKEN_BLANK: parse_blank,
    TOKEN_STRING: parse_string_literal,
    TOKEN_INTEGER: parse_integer_literal,
    TOKEN_FLOAT: parse_float_literal,
    TOKEN_IDENTIFIER: parse_identifier,
}

BINARY_OPERATORS = frozenset(
    (
        TOKEN_EQ,
        TOKEN_OR,
        TOKEN_AND,
        TOKEN_LT,
        TOKEN_GT,
        TOKEN_NE,
        TOKEN_LG,
        TOKEN_LE,
        TOKEN_GE,
        TOKEN_CONTAINS,
    )
)

PREFIX_OPERATORS = frozenset(
    [
        TOKEN_NOT,
    ]
)


def parse_prefix_expression(stream: TokenStream) -> Expression:
    """Parse a prefix expression from a stream of tokens.

    Note that "standard" Liquid boolean expressions do not have any prefix expressions.
    """
    tok = next(stream)
    assert tok[2] == TOKEN_NOT
    return PrefixExpression(
        operator=tok[2],
        right=parse_obj_with_parens(
            stream,
            precedence=PRECEDENCE_LOGICALRIGHT,
        ),
    )


def parse_infix_expression(stream: TokenStream, left: Expression) -> InfixExpression:
    """Parse an infix expression from a stream of tokens."""
    tok = stream.current
    precedence = PRECEDENCES.get(tok[1], PRECEDENCE_LOWEST)
    stream.next_token()

    exp = InfixExpression(
        left=left,
        operator=tok[2],
        right=parse_obj(stream, precedence),
    )
    return exp


def parse_infix_expression_with_parens(
    stream: TokenStream,
    left: Expression,
) -> InfixExpression:
    """Parse an infix expression from a stream of tokens."""
    tok = stream.current
    precedence = PRECEDENCES.get(tok[1], PRECEDENCE_LOWEST)
    stream.next_token()

    exp = InfixExpression(
        left=left,
        operator=tok[2],
        right=parse_obj_with_parens(stream, precedence),
    )
    return exp


def parse_obj(
    stream: TokenStream,
    precedence: int = PRECEDENCE_LOWEST,
) -> Expression:
    """Parse the next object from the stream of tokens.

    This object parser is for "standard" boolean expressions. It does not
    handle logical `not` or grouping with parentheses.
    """
    try:
        left = TOKEN_MAP[stream.current[1]](stream)
    except KeyError as err:
        raise LiquidSyntaxError(
            f"unexpected {stream.current[2]!r}",
            linenum=stream.current[0],
        ) from err

    while True:
        peek_typ = stream.peek[1]
        if (
            peek_typ == TOKEN_EOF
            or PRECEDENCES.get(peek_typ, PRECEDENCE_LOWEST) < precedence
        ):
            break

        if peek_typ not in BINARY_OPERATORS:
            return left

        next(stream)
        left = parse_infix_expression(stream, left)

    return left


parse_range = make_parse_range(parse_simple_obj)
TOKEN_MAP[TOKEN_LPAREN] = parse_range


def parse(expr: str, linenum: int = 1) -> BooleanExpression:
    """Parse a string as a "standard" boolean expression."""
    return BooleanExpression(parse_obj(TokenStream(tokenize(expr, linenum))))


def parse_grouped_expression(stream: TokenStream) -> Expression:
    """Parse a group of logical expressions."""
    next(stream)  # Eat left paren
    expr = parse_obj_with_parens(stream)

    next(stream)
    while stream.current[1] == TOKEN_RPAREN:
        next(stream)

    if stream.current[1] != TOKEN_EOF:
        expr = parse_infix_expression_with_parens(stream, left=expr)

    return expr


def parse_range_with_parens(stream: TokenStream) -> Expression:
    """Like `parse_range` but consumes the extra `RANGE_LPAREN` token first."""
    stream.expect(TOKEN_RANGE_LITERAL)
    next(stream)  # Eat extra token
    return parse_range(stream)


TOKEN_MAP_WITH_PARENS = {
    **TOKEN_MAP,
    TOKEN_NOT: parse_prefix_expression,
    TOKEN_LPAREN: parse_grouped_expression,
    TOKEN_RANGE_LITERAL: parse_range_with_parens,
}


def parse_obj_with_parens(
    stream: TokenStream,
    precedence: int = PRECEDENCE_LOWEST,
) -> Expression:
    """Parse the next object from the stream of tokens.

    This object parser is for the non-standard boolean expression, which does handle
    the logical `not` operator and grouping terms with parentheses.
    """
    try:
        left = TOKEN_MAP_WITH_PARENS[stream.current[1]](stream)
    except KeyError as err:
        raise LiquidSyntaxError(
            f"unexpected {stream.current[2]!r}",
            linenum=stream.current[0],
        ) from err

    while True:
        peek_typ = stream.peek[1]
        if (
            peek_typ == TOKEN_EOF
            or PRECEDENCES.get(peek_typ, PRECEDENCE_LOWEST) < precedence
        ):
            break

        if peek_typ not in BINARY_OPERATORS:
            return left

        next(stream)
        left = parse_infix_expression_with_parens(stream, left)

    return left


def parse_with_parens(expr: str, linenum: int = 1) -> BooleanExpression:
    """Parse a string as a boolean expression, possibly containing the logical `not`
    operator and parentheses for grouping terms.
    """
    return BooleanExpression(
        parse_obj_with_parens(
            TokenStream(
                tokenize_with_parens(expr, linenum),
            )
        )
    )



# pylint: disable=missing-module-docstring
from __future__ import annotations
from typing import TYPE_CHECKING

from .filters import index
from .filters import JSON
from .filters import script_tag
from .filters import stylesheet_tag

from .tags import CallTag
from .tags import IfNotTag
from .tags import InlineIfAssignTag
from .tags import InlineIfAssignTagWithParens
from .tags import InlineIfEchoTag
from .tags import InlineIfEchoTagWithParens
from .tags import InlineIfStatement
from .tags import InlineIfStatementWithParens
from .tags import MacroTag
from .tags import WithTag


if TYPE_CHECKING:  # pragma: no cover
    from liquid import Environment

__all__ = (
    "index",
    "JSON",
    "script_tag",
    "stylesheet_tag",
    "CallTag",
    "IfNotTag",
    "InlineIfAssignTag",
    "InlineIfAssignTagWithParens",
    "InlineIfEchoTag",
    "InlineIfEchoTagWithParens",
    "InlineIfStatement",
    "InlineIfStatementWithParens",
    "MacroTag",
    "WithTag",
    "add_filters",
    "add_inline_expression_tags",
    "add_extended_inline_expression_tags",
    "add_macro_tags",
    "add_tags",
    "add_tags_and_filters",
)


def add_inline_expression_tags(env: Environment) -> None:
    """Replace standard implementations of the output statement,
    `echo` tag and `assign` tag with ones that support inline `if`
    expressions."""
    env.add_tag(InlineIfAssignTag)
    env.add_tag(InlineIfEchoTag)
    env.add_tag(InlineIfStatement)


def add_extended_inline_expression_tags(env: Environment) -> None:
    """Replace standard implementations of the output statement,
    `echo` tag and `assign` tag with ones that support inline `if`
    expressions."""
    env.add_tag(InlineIfAssignTagWithParens)
    env.add_tag(InlineIfEchoTagWithParens)
    env.add_tag(InlineIfStatementWithParens)


def add_macro_tags(env: Environment) -> None:
    """Register both the `macro` and `call` tags with an environment."""
    env.add_tag(CallTag)
    env.add_tag(MacroTag)


def add_tags(env: Environment) -> None:  # pragma: no cover
    """Register all extra tags with an environment."""
    env.add_tag(IfNotTag)
    env.add_tag(WithTag)
    add_macro_tags(env)
    add_extended_inline_expression_tags(env)


def add_filters(env: Environment) -> None:
    """Register all extra filters with an environment with their default
    and options."""
    env.add_filter("index", index)
    env.add_filter("json", JSON())
    env.add_filter("script_tag", script_tag)
    env.add_filter("stylesheet_tag", stylesheet_tag)


def add_tags_and_filters(env: Environment) -> None:  # pragma: no cover
    """Register all extra tags and filters with an environment."""
    add_tags(env)
    add_filters(env)



"""A wrapper for token iterators that lets us step through and peek ahead."""

from __future__ import annotations

from collections import deque

from typing import Deque
from typing import Iterator
from typing import List

from liquid.token import Token
from liquid.token import TOKEN_INITIAL
from liquid.token import TOKEN_EOF


class TokenStream:
    """Step through or iterate a stream of tokens."""

    def __init__(self, tokeniter: Iterator[Token]):
        self.iter = tokeniter

        # Queue of peeked tokens
        self._pushed: Deque[Token] = deque()

        # Stack of tags
        self.balancing_stack: List[str] = []

        self.current: Token = Token(0, TOKEN_INITIAL, "")
        next(self)

    class TokenStreamIterator:
        """An iterable token stream."""

        def __init__(self, stream: TokenStream):
            self.stream = stream

        def __iter__(self) -> Iterator[Token]:
            return self

        def __next__(self) -> Token:
            tok = self.stream.current
            if tok.type is TOKEN_EOF:
                self.stream.close()
                raise StopIteration()
            next(self.stream)
            return tok

    def __iter__(self) -> Iterator[Token]:
        return self.TokenStreamIterator(self)

    def __next__(self) -> Token:
        tok = self.current
        if self._pushed:
            self.current = self._pushed.popleft()
        elif self.current.type is not TOKEN_EOF:
            try:
                self.current = next(self.iter)
            except StopIteration:
                self.close()
        return tok

    def __str__(self) -> str:  # pragma: no cover
        buf = [
            f"current: {self.current}",
            f"next: {self.peek}",
        ]
        return "\n".join(buf)

    def next_token(self) -> Token:
        """Return the next token from the stream."""
        return next(self)

    @property
    def peek(self) -> Token:
        """Look at the next token."""
        current = next(self)
        result = self.current
        self.push(current)
        return result

    def push(self, tok: Token) -> None:
        """Push a token back to the stream."""
        self._pushed.append(self.current)
        self.current = tok

    def close(self) -> None:
        """Close the stream."""
        self.current = Token(0, TOKEN_EOF, "")



"""Golden tests cases for testing liquid's built-in `replace` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="replace substrings",
        template=(
            r'{{ "Take my protein pills and put my helmet on" '
            r'| replace: "my", "your" }}'
        ),
        expect="Take your protein pills and put your helmet on",
    ),
    Case(
        description="not a string",
        template=r"{{ 5 | replace: 'rain', 'foo' }}",
        expect="5",
    ),
    Case(
        description="left value is an object",
        template=r"{{ a | replace: '{', '!' }}",
        expect="!}",
        globals={"a": {}},
    ),
    Case(
        description="argument not a string",
        template=r'{{ "hello" | replace: 5, "your" }}',
        expect="hello",
    ),
    Case(
        description="missing argument",
        template=r'{{ "hello" | replace: "ll" }}',
        expect="heo",
    ),
    Case(
        description="missing arguments",
        template=r'{{ "hello" | replace }}',
        expect="",
        error=True,
    ),
    Case(
        description="too many arguments",
        template=r'{{ "hello" | replace: "how", "are", "you" }}',
        expect="",
        error=True,
    ),
    Case(
        description="undefined left value",
        template=r'{{ nosuchthing | replace: "my", "your" }}',
        expect="",
    ),
    Case(
        description="undefined first argument",
        template=r'{{ "Take my protein" | replace: nosuchthing, "#" }}',
        expect="#T#a#k#e# #m#y# #p#r#o#t#e#i#n#",
    ),
    Case(
        description="undefined second argument",
        template=(
            r'{{ "Take my protein pills and put my helmet on" '
            r'| replace: "my", nosuchthing }}'
        ),
        expect="Take  protein pills and put  helmet on",
    ),
]



"""Golden tests cases for testing liquid's built-in `strip_html` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="some HTML markup",
        template=r"{{ s | strip_html }}",
        expect="Have you read Ulysses &amp; &#20;?",
        globals={"s": "Have <em>you</em> read <strong>Ulysses</strong> &amp; &#20;?"},
    ),
    Case(
        description="some HTML markup with HTML comment",
        template=r"{{ s | strip_html }}",
        expect="you read Ulysses &amp; &#20;?",
        globals={
            "s": "<!-- Have --><em>you</em> read <strong>Ulysses</strong> &amp; &#20;?"
        },
    ),
    Case(
        description="not a string",
        template=r"{{ 5 | strip_html }}",
        expect="5",
    ),
    Case(
        description="unexpected argument",
        template=r'{{ "hello" | strip_html: 5 }}',
        expect="",
        error=True,
    ),
    Case(
        description="undefined left value",
        template=r"{{ nosuchthing | strip_html }}",
        expect="",
    ),
    Case(
        description="html block",
        template=r"{{ s | strip_html }}",
        expect="test",
        globals={"s": "<div>test</div>"},
    ),
    Case(
        description="html block with id",
        template=r"{{ s | strip_html }}",
        expect="test",
        globals={"s": "<div id='test'>test</div>"},
    ),
    Case(
        description="script block",
        template=r"{{ s | strip_html }}",
        expect="",
        globals={
            "s": "<script type='text/javascript'>document.write('some stuff');</script>"
        },
    ),
    Case(
        description="style block",
        template=r"{{ s | strip_html }}",
        expect="",
        globals={"s": "<style type='text/css'>foo bar</style>"},
    ),
    Case(
        description="html block with newline",
        template=r"{{ s | strip_html }}",
        expect="test",
        globals={"s": "<div\nclass='multiline'>test</div>"},
    ),
    Case(
        description="html comment with newline",
        template=r"{{ s | strip_html }}",
        expect="test",
        globals={"s": "<!-- foo bar \n test -->test"},
    ),
]



# pylint: disable=missing-module-docstring
from __future__ import annotations

from typing import TYPE_CHECKING

from . import literal
from . import statement
from . import illegal

from .tags import if_tag
from .tags import comment_tag
from .tags import unless_tag
from .tags import case_tag
from .tags import for_tag
from .tags import tablerow_tag
from .tags import capture_tag
from .tags import cycle_tag
from .tags import assign_tag
from .tags import increment_tag
from .tags import decrement_tag
from .tags import echo_tag
from .tags import liquid_tag
from .tags import include_tag
from .tags import render_tag
from .tags import ifchanged_tag
from .tags import inline_comment_tag

from .filters.math import abs_
from .filters.math import at_most
from .filters.math import at_least
from .filters.math import ceil
from .filters.math import divided_by
from .filters.math import floor
from .filters.math import minus
from .filters.math import plus
from .filters.math import round_
from .filters.math import times
from .filters.math import modulo

from .filters.string import capitalize
from .filters.string import append
from .filters.string import downcase
from .filters.string import escape
from .filters.string import escape_once
from .filters.string import lstrip
from .filters.string import newline_to_br
from .filters.string import prepend
from .filters.string import remove
from .filters.string import remove_first
from .filters.string import remove_last
from .filters.string import replace
from .filters.string import replace_first
from .filters.string import replace_last
from .filters.string import slice_
from .filters.string import split
from .filters.string import upcase
from .filters.string import strip
from .filters.string import rstrip
from .filters.string import strip_html
from .filters.string import strip_newlines
from .filters.string import truncate
from .filters.string import truncatewords
from .filters.string import url_encode
from .filters.string import url_decode
from .filters.string import base64_encode
from .filters.string import base64_decode
from .filters.string import base64_url_safe_encode
from .filters.string import base64_url_safe_decode

from .filters.array import join
from .filters.array import first
from .filters.array import last
from .filters.array import concat
from .filters.array import map_
from .filters.array import reverse
from .filters.array import sort
from .filters.array import sort_natural
from .filters.array import where
from .filters.array import uniq
from .filters.array import compact

from .filters.misc import size
from .filters.misc import default
from .filters.misc import date

from .filters.extra import safe


if TYPE_CHECKING:
    from liquid import Environment


# pylint: disable=too-many-statements
def register(env: Environment) -> None:
    """Register all built-in tags and filters with an environment."""
    env.add_tag(literal.Literal)
    env.add_tag(statement.Statement)
    env.add_tag(illegal.Illegal)
    env.add_tag(if_tag.IfTag)
    env.add_tag(comment_tag.CommentTextTag)
    env.add_tag(unless_tag.UnlessTag)
    env.add_tag(case_tag.CaseTag)
    env.add_tag(for_tag.ForTag)
    env.add_tag(for_tag.BreakTag)
    env.add_tag(for_tag.ContinueTag)
    env.add_tag(tablerow_tag.TablerowTag)
    env.add_tag(capture_tag.CaptureTag)
    env.add_tag(cycle_tag.CycleTag)
    env.add_tag(assign_tag.AssignTag)
    env.add_tag(increment_tag.IncrementTag)
    env.add_tag(decrement_tag.DecrementTag)
    env.add_tag(echo_tag.EchoTag)
    env.add_tag(liquid_tag.LiquidTag)
    env.add_tag(include_tag.IncludeTag)
    env.add_tag(render_tag.RenderTag)
    env.add_tag(ifchanged_tag.IfChangedTag)
    env.add_tag(inline_comment_tag.InlineCommentTag)

    env.add_filter("abs", abs_)
    env.add_filter("at_most", at_most)
    env.add_filter("at_least", at_least)
    env.add_filter("ceil", ceil)
    env.add_filter("divided_by", divided_by)
    env.add_filter("floor", floor)
    env.add_filter("minus", minus)
    env.add_filter("plus", plus)
    env.add_filter("round", round_)
    env.add_filter("times", times)
    env.add_filter("modulo", modulo)

    env.add_filter("capitalize", capitalize)
    env.add_filter("append", append)
    env.add_filter("downcase", downcase)
    env.add_filter("escape", escape)
    env.add_filter("escape_once", escape_once)
    env.add_filter("lstrip", lstrip)
    env.add_filter("newline_to_br", newline_to_br)
    env.add_filter("prepend", prepend)
    env.add_filter("remove", remove)
    env.add_filter("remove_first", remove_first)
    env.add_filter("remove_last", remove_last)
    env.add_filter("replace", replace)
    env.add_filter("replace_first", replace_first)
    env.add_filter("replace_last", replace_last)
    env.add_filter("slice", slice_)
    env.add_filter("split", split)
    env.add_filter("upcase", upcase)
    env.add_filter("strip", strip)
    env.add_filter("rstrip", rstrip)
    env.add_filter("strip_html", strip_html)
    env.add_filter("strip_newlines", strip_newlines)
    env.add_filter("truncate", truncate)
    env.add_filter("truncatewords", truncatewords)
    env.add_filter("url_encode", url_encode)
    env.add_filter("url_decode", url_decode)
    env.add_filter("base64_encode", base64_encode)
    env.add_filter("base64_decode", base64_decode)
    env.add_filter("base64_url_safe_encode", base64_url_safe_encode)
    env.add_filter("base64_url_safe_decode", base64_url_safe_decode)

    env.add_filter("join", join)
    env.add_filter("first", first)
    env.add_filter("last", last)
    env.add_filter("concat", concat)
    env.add_filter("map", map_)
    env.add_filter("reverse", reverse)
    env.add_filter("sort", sort)
    env.add_filter("sort_natural", sort_natural)
    env.add_filter("where", where)
    env.add_filter("uniq", uniq)
    env.add_filter("compact", compact)

    env.add_filter("size", size)
    env.add_filter("default", default)
    env.add_filter("date", date)

    env.add_filter("safe", safe)



"""Golden tests cases for testing liquid's built-in `strip` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="left padded",
        template='{{ " \t\r\n  hello" | strip }}',
        expect="hello",
    ),
    Case(
        description="right padded",
        template='{{ "hello \t\r\n  " | strip }}',
        expect="hello",
    ),
    Case(
        description="left and right padded",
        template='{{ " \t\r\n  hello  \t\r\n " | strip }}',
        expect="hello",
    ),
    Case(
        description="not a string",
        template=r"{{ 5 | strip }}",
        expect="5",
    ),
    Case(
        description="unexpected argument",
        template=r'{{ "hello" | strip: 5 }}',
        expect="",
        error=True,
    ),
    Case(
        description="undefined left value",
        template=r"{{ nosuchthing | strip }}",
        expect="",
    ),
]



"""Golden tests cases for testing liquid's built-in `minus` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="integer value and integer arg",
        template=r"{{ 10 | minus: 2 }}",
        expect="8",
    ),
    Case(
        description="integer value and float arg",
        template=r"{{ 10 | minus: 2.0 }}",
        expect="8.0",
    ),
    Case(
        description="float value and float arg",
        template=r"{{ 10.1 | minus: 2.2 }}",
        expect="7.9",
    ),
    Case(
        description="string value and string arg",
        template=r'{{ "10.1" | minus: "2.2" }}',
        expect="7.9",
    ),
    Case(
        description="string not a number",
        template=r'{{ "foo" | minus: "2.0" }}',
        expect="-2.0",
    ),
    Case(
        description="arg string not a number",
        template=r'{{ "10" | minus: "foo" }}',
        expect="10",
    ),
    Case(
        description="too many args",
        template=r"{{ 5 | minus: 1, '5' }}",
        expect="",
        error=True,
    ),
    Case(
        description="not a string, int or float",
        template=r"{{ a | minus: 1 }}",
        expect="-1",
        globals={"a": {}},
    ),
    Case(
        description="undefined left value",
        template=r"{{ nosuchthing | minus: 2 }}",
        expect="-2",
    ),
    Case(
        description="undefined argument",
        template=r"{{ 10 | minus: nosuchthing }}",
        expect="10",
    ),
]



"""Golden tests cases for testing liquid's `capture` tag."""

from liquid.golden.case import Case

cases = [
    Case(
        description="capture template literal and global variable",
        template=(
            r"{% capture greeting %}"
            r"Hello, {{ customer.first_name }}."
            r"{% endcapture %}"
            r"{{ greeting }}"
        ),
        expect="Hello, Holly.",
        globals={"customer": {"first_name": "Holly"}},
    ),
    Case(
        description="capture into a variable with a hyphen",
        template=(
            r"{% capture this-thing %}"
            r"Hello, {{ customer.first_name }}."
            r"{% endcapture %}"
            r"{{ this-thing }}"
        ),
        expect="Hello, Holly.",
        globals={"customer": {"first_name": "Holly"}},
    ),
    Case(
        description="assign to a variable from a captured variable",
        template=(
            r"{% capture some %}"
            r"hello"
            r"{% endcapture %}"
            r"{% assign other = some %}"
            r"{{ some }}-{{ other }}"
        ),
        expect="hello-hello",
    ),
]



"""A drop-in replacement for the standard `if` tag that supports the
logical `not` operator and grouping terms with parentheses."""

from liquid.builtin.tags.if_tag import IfTag
from liquid.expression import Expression
from liquid.expressions import parse_boolean_expression_with_parens
from liquid.parse import expect
from liquid.stream import TokenStream
from liquid.token import TOKEN_EXPRESSION


class IfNotTag(IfTag):
    """A drop-in replacement for the standard `if` tag that supports the
    logical `not` operator and grouping terms with parentheses."""

    def parse_expression(self, stream: TokenStream) -> Expression:
        """Pare a boolean expression from a stream of tokens."""
        expect(stream, TOKEN_EXPRESSION)
        return parse_boolean_expression_with_parens(stream.current.value)



"""Mock HTML tag related filter functions."""
# pylint: disable=missing-function-docstring
from liquid.filter import string_filter
from liquid.filter import liquid_filter
from liquid.filter import with_context


@liquid_filter
@with_context
def link_to_tag(label, tag, *, context):
    handle = context.resolve("handle", default="")
    return f'<a title="Show tag {tag}" href="/collections/{handle}/{tag}">{label}</a>'


@string_filter
def highlight_active_tag(tag, css_class="active"):
    return f'<span class="{css_class}">{tag}</span>'


@liquid_filter
@with_context
def link_to_add_tag(label, tag, *, context):
    handle = context.resolve("handle", default="")
    tags = "+".join(set(context.resolve("current_tags", default=[]) + [tag]))
    return f'<a title="Show tag {tag}" href="/collections/{handle}/{tags}">{label}</a>'


@liquid_filter
@with_context
def link_to_remove_tag(label, tag, *, context):
    handle = context.resolve("handle", default="")
    current_tags = set(context.resolve("current_tags", default=[]))
    tags = "+".join(t for t in current_tags if t != tag)
    return f'<a title="Hide tag {tag}" href="/collections/{handle}/{tags}">{label}</a>'



"""Patterns and parse functions common to multiple built-in expression types."""
import sys

from typing import Callable
from typing import Tuple
from typing import Union
from typing import TYPE_CHECKING

from liquid.expression import Expression
from liquid.expression import TRUE
from liquid.expression import FALSE
from liquid.expression import BLANK
from liquid.expression import NIL
from liquid.expression import EMPTY
from liquid.expression import Nil
from liquid.expression import Blank
from liquid.expression import Empty
from liquid.expression import Boolean
from liquid.expression import Identifier
from liquid.expression import IdentifierPath
from liquid.expression import IdentifierPathElement
from liquid.expression import StringLiteral
from liquid.expression import IntegerLiteral
from liquid.expression import FloatLiteral
from liquid.expression import RangeLiteral

from liquid.token import reverse_operators
from liquid.token import TOKEN_TRUE
from liquid.token import TOKEN_IDENTIFIER
from liquid.token import TOKEN_IDENTINDEX
from liquid.token import TOKEN_DOT
from liquid.token import TOKEN_LBRACKET
from liquid.token import TOKEN_RBRACKET
from liquid.token import TOKEN_INTEGER
from liquid.token import TOKEN_FLOAT
from liquid.token import TOKEN_LPAREN
from liquid.token import TOKEN_RPAREN
from liquid.token import TOKEN_RANGE
from liquid.token import TOKEN_STRING

from liquid.exceptions import LiquidSyntaxError
from liquid.limits import to_int

if TYPE_CHECKING:
    from liquid.expressions.stream import TokenStream

GROUP_QUOTE = sys.intern("quote")
GROUP_QUOTED = sys.intern("quoted")
GROUP_IDENTINDEX = sys.intern("path_index")
GROUP_IDENTQUOTE = sys.intern("identquote")
GROUP_IDENTQUOTED = sys.intern("identquoted")

# The reference implementation will accept identifiers that start with an integer or
# hyphen when in lax mode, but will raise a syntax error in strict mode. For historic
# reasons we'll stick the following pattern until at least version 2 of Python Liquid.
IDENTIFIER_PATTERN = r"\w[\w\-]*\??"

# Trailing ? is not allowed in assignment names
ASSIGN_IDENTIFIER_PATTERN = r"\w[\w\-]*"

# ["ident"] or ['ident']
IDENTSTRING_PATTERN = (
    rf"\[\s*(?P<{GROUP_IDENTQUOTE}>[\"'])"
    rf"(?P<{GROUP_IDENTQUOTED}>.*?)"
    rf"(?P={GROUP_IDENTQUOTE})\s*]"
)

# [0] or [-1]
IDENTINDEX_PATTERN = rf"\[\s*(?P<{GROUP_IDENTINDEX}>\-?\d+)\s*]"

# 'something' or "something"
STRING_PATTERN = rf"(?P<{GROUP_QUOTE}>[\"'])(?P<{GROUP_QUOTED}>.*?)(?P={GROUP_QUOTE})"

# (position, type, value)
Token = Tuple[int, str, str]


def parse_boolean(stream: "TokenStream") -> Boolean:
    """Read a boolean literal (true or false) from the token stream."""
    if stream.current[1] == TOKEN_TRUE:
        return TRUE
    return FALSE


def parse_nil(_: "TokenStream") -> Nil:
    """Read a 'nil' keyword from the token stream."""
    return NIL


def parse_empty(_: "TokenStream") -> Empty:
    """Read a 'empty' keyword from the token stream."""
    return EMPTY


def parse_blank(_: "TokenStream") -> Blank:
    """Read a 'blank' keyword from the token stream."""
    return BLANK


def parse_string_literal(stream: "TokenStream") -> StringLiteral:
    """Read a string from the token stream."""
    return StringLiteral(value=stream.current[2])


def parse_integer_literal(stream: "TokenStream") -> IntegerLiteral:
    """Read an integer from the token stream."""
    return IntegerLiteral(value=to_int(stream.current[2]))


def parse_float_literal(stream: "TokenStream") -> FloatLiteral:
    """Read a float from the token stream."""
    return FloatLiteral(value=float(stream.current[2]))


IDENT_TOKENS = frozenset(
    (
        TOKEN_IDENTIFIER,
        TOKEN_IDENTINDEX,
        TOKEN_LBRACKET,
        TOKEN_DOT,
    )
)


def parse_identifier(stream: "TokenStream") -> Identifier:
    """Read an identifier from the token stream.

    An identifier might be chained with dots and square brackets, and might contain
    more, possibly chained, identifiers within those brackets.
    """
    path: IdentifierPath = []
    stream.expect(TOKEN_IDENTIFIER)

    while True:
        pos, typ, val = stream.current
        if typ == TOKEN_IDENTIFIER:
            path.append(IdentifierPathElement(val))
        elif typ == TOKEN_IDENTINDEX:
            path.append(IdentifierPathElement(to_int(val)))
        elif typ == TOKEN_LBRACKET:
            stream.next_token()
            path.append(parse_identifier(stream))
            # Eat close bracket
            stream.next_token()
            stream.expect(TOKEN_RBRACKET)
        elif typ == TOKEN_FLOAT:
            raise LiquidSyntaxError(
                f"expected and identifier, found {val!r}",
                linenum=pos,
            )
        elif typ == TOKEN_DOT:
            pass
        else:
            stream.push(stream.current)
            break

        stream.next_token()

    return Identifier(path)


def parse_string_or_identifier(
    stream: "TokenStream",
) -> Union[StringLiteral, Identifier]:
    """Parse an expression from a stream of tokens. If the stream is not at a string or
    identifier expression, raise a syntax error.
    """
    typ = stream.current[1]
    if typ == TOKEN_IDENTIFIER:
        expr: Union[StringLiteral, Identifier] = parse_identifier(stream)
    elif typ == TOKEN_STRING:
        expr = parse_string_literal(stream)
    else:
        _typ = reverse_operators.get(typ, typ)
        msg = f"expected identifier or string, found {_typ}"
        raise LiquidSyntaxError(msg, linenum=stream.current[0])

    return expr


def parse_unchained_identifier(stream: "TokenStream") -> Identifier:
    """Parse an identifier from a stream of tokens. If the stream is not at an
    identifier or the identifier is chained, raise a syntax error."""
    tok = stream.current
    ident = parse_identifier(stream)
    if len(ident.path) != 1:
        raise LiquidSyntaxError(f"invalid identifier '{ident}'", linenum=tok[0])
    return ident


def make_parse_range(
    parse_obj: Callable[["TokenStream"], Expression]
) -> Callable[["TokenStream"], RangeLiteral]:
    """Return a function that parse range expressions using the given parse_obj
    callable.
    """

    def _parse_range_literal(stream: "TokenStream") -> RangeLiteral:
        """Read a range literal from the token stream.

        A range literal can contain an identifier (possibly chained), an integer or a
        float as its start and stop values.
        """
        # Eat left parenthesis
        stream.expect(TOKEN_LPAREN)
        stream.next_token()

        # Parse start
        if stream.current[1] not in (TOKEN_IDENTIFIER, TOKEN_INTEGER, TOKEN_FLOAT):
            raise LiquidSyntaxError(
                f"unexpected {stream.current[2]!r} in range expression",
                linenum=stream.current[0],
            )
        start = parse_obj(stream)
        stream.next_token()

        # Eat TOKEN_RANGE
        stream.expect(TOKEN_RANGE)
        stream.next_token()

        # Parse stop
        if stream.current[1] not in (TOKEN_IDENTIFIER, TOKEN_INTEGER, TOKEN_FLOAT):
            raise LiquidSyntaxError(
                f"unexpected {stream.current[2]!r} in range expression",
                linenum=stream.current[0],
            )
        stop = parse_obj(stream)

        expr = RangeLiteral(start, stop)

        # Eat right parenthesis
        stream.next_token()
        stream.expect(TOKEN_RPAREN)

        return expr

    return _parse_range_literal



"""Drop-in replacements for the standard output statement, `assign` tag
and `echo` tag that support inline `if` expressions, optionally including
the logical `not` operator and grouping terms with parentheses."""

from liquid.expression import Expression
from liquid.parse import expect
from liquid.stream import TokenStream
from liquid.token import TOKEN_STATEMENT

from liquid.builtin.tags.assign_tag import AssignTag
from liquid.builtin.tags.echo_tag import EchoTag
from liquid.builtin.statement import Statement
from liquid.builtin.statement import StatementNode


class InlineIfStatement(Statement):
    """A drop-in replacement for the standard output statement that supports
    inline ``if`` expressions."""

    def parse(self, stream: TokenStream) -> StatementNode:
        tok = stream.current
        expect(stream, TOKEN_STATEMENT)
        return StatementNode(
            tok, self.env.parse_conditional_expression_value(tok.value)
        )


class InlineIfAssignTag(AssignTag):
    """A drop-in replacement for the standard ``assign`` tag that supports
    inline ``if`` expressions."""

    def _parse_expression(self, value: str) -> Expression:
        return self.env.parse_conditional_expression_value(value)


class InlineIfEchoTag(EchoTag):
    """A drop-in replacement for the standard ``echo`` tag that supports
    inline ``if`` expressions."""

    def _parse_expression(self, value: str) -> Expression:
        return self.env.parse_conditional_expression_value(value)


class InlineIfStatementWithParens(Statement):
    """A drop-in replacement for the standard output statement that supports
    inline ``if`` expressions with the logical ``not`` operator and grouping
    terms with parentheses."""

    def parse(self, stream: TokenStream) -> StatementNode:
        tok = stream.current
        expect(stream, TOKEN_STATEMENT)
        return StatementNode(
            tok, self.env.parse_conditional_expression_value_with_parens(tok.value)
        )


class InlineIfAssignTagWithParens(AssignTag):
    """A drop-in replacement for the standard ``assign`` tag that supports
    inline ``if`` expressions with the logical ``not`` operator and grouping
    terms with parentheses."""

    def _parse_expression(self, value: str) -> Expression:
        return self.env.parse_conditional_expression_value_with_parens(value)


class InlineIfEchoTagWithParens(EchoTag):
    """A drop-in replacement for the standard ``echo`` tag that supports
    inline ``if`` expressions with the logical ``not`` operator and grouping
    terms with parentheses."""

    def _parse_expression(self, value: str) -> Expression:
        return self.env.parse_conditional_expression_value_with_parens(value)



# pylint: disable=missing-class-docstring missing-module-docstring
import unittest

from typing import Any
from typing import Mapping
from typing import NamedTuple

from liquid import Context
from liquid import Environment

from liquid.context import Undefined

from liquid.expressions import parse_filtered_expression
from liquid.expressions import parse_conditional_expression
from liquid.expressions import parse_conditional_expression_with_parens


class Case(NamedTuple):
    """Table-driven test case helper."""

    description: str
    context: Mapping[str, Any]
    expression: str
    expect: Any


class EvalFilteredExpressionTestCase(unittest.TestCase):
    """Test cases for evaluating standard filtered expressions."""

    test_cases = [
        Case(
            "string literal",
            {},
            "'foobar'",
            "foobar",
        ),
        Case(
            "integer literal",
            {},
            "7",
            7,
        ),
        Case(
            "float literal",
            {},
            "7.5",
            7.5,
        ),
        Case(
            "negative integer literal",
            {},
            "-7",
            -7,
        ),
        Case(
            "negative float literal",
            {},
            "-7.5",
            -7.5,
        ),
        Case(
            "single global object identifier",
            {"collection": "foo"},
            "collection",
            "foo",
        ),
        Case(
            "string literal with no arg filter",
            {},
            "'foo' | upcase",
            "FOO",
        ),
        Case(
            "object identifier with no arg filter",
            {"collection": {"title": "foo"}},
            "collection.title | upcase",
            "FOO",
        ),
        Case(
            "string literal with two arg filter",
            {},
            '"Liquid" | slice: 2, 5',
            "quid",
        ),
        Case(
            "string literal with two filters",
            {},
            '"Liquid" | slice: 2, 5 | upcase',
            "QUID",
        ),
        Case(
            "resolve identifier chain",
            {"a": {"b": {"c": "hello", "array": [1, 2, 3]}}},
            "a.b.c",
            "hello",
        ),
        Case(
            "resolve identifier chain not in context",
            {"a": {"b": {"c": "hello", "array": [1, 2, 3]}}},
            "a.b.x",
            None,
        ),
        Case(
            "resolve identifier chain containing whitespace.",
            {"a": {"b x": {"c": "hello", "array": [1, 2, 3]}}},
            "a['b x'].c",
            "hello",
        ),
        Case(
            "resolve identifier chain ending in an array",
            {"a": {"b": {"c": "hello", "array": [1, 2, 3]}}},
            "a.b.array",
            [1, 2, 3],
        ),
        Case(
            "resolve identifier chain ending in an array index using subscript",
            {"a": {"b": {"c": "hello", "array": [1, 2, 3]}}},
            "a.b.array[1]",
            2,
        ),
        Case(
            "try to read past an array",
            {"a": {"b": {"c": "hello", "array": [1, 2, 3]}}},
            "a.b.array.foo",
            None,
        ),
        Case(
            "array `first` special method",
            {"a": {"b": {"c": "hello", "array": [1, 2, 3]}}},
            "a.b.array.first",
            1,
        ),
        Case(
            "array `last` special method",
            {"a": {"b": {"c": "hello", "array": [1, 2, 3]}}},
            "a.b.array.last",
            3,
        ),
        Case(
            "array `size` special method",
            {"a": {"b": {"c": "hello", "array": [1, 2, 3]}}},
            "a.b.array.size",
            3,
        ),
        Case(
            "size of an empty array",
            {"a": {"b": {"c": "hello", "array": []}}},
            "a.b.array.size",
            0,
        ),
        Case(
            "size of an object",
            {"a": {"b": {"c": "hello", "array": []}}},
            "a.b.size",
            2,
        ),
        Case(
            "nested and chained",
            {
                "linklists": {"main": "main menu"},
                "section": {"settings": {"menu": "main"}},
            },
            "linklists[section.settings.menu]",
            "main menu",
        ),
        Case(
            "array index using negative subscript",
            {"a": [1, 2, 3]},
            "a[-1]",
            3,
        ),
    ]

    def test_eval_filtered_expression(self) -> None:
        """Test that we can evaluate standard boolean expressions."""
        env = Environment()

        for case in self.test_cases:
            with self.subTest(msg=case.description):
                context = Context(env, globals=case.context)
                expr = parse_filtered_expression(case.expression)
                self.assertEqual(expr.evaluate(context), case.expect)

    def test_eval_conditional_expression(self) -> None:
        """Test that we can evaluation of non-standard boolean expressions
        is backwards compatible."""
        env = Environment()

        for case in self.test_cases:
            with self.subTest(msg=case.description):
                context = Context(env, globals=case.context)
                expr = parse_conditional_expression(case.expression)
                self.assertEqual(expr.evaluate(context), case.expect)

    def test_eval_conditional_expression_with_parens(self) -> None:
        """Test that we can evaluation of non-standard boolean expressions
        is backwards compatible."""
        env = Environment()

        for case in self.test_cases:
            with self.subTest(msg=case.description):
                context = Context(env, globals=case.context)
                expr = parse_conditional_expression_with_parens(case.expression)
                self.assertEqual(expr.evaluate(context), case.expect)


class EvalConditionalExpressionTestCase(unittest.TestCase):
    """Test cases for evaluating non-standard conditional expressions."""

    test_cases = [
        Case(
            description="string literal with true condition",
            context={},
            expression="'foo' if true",
            expect="foo",
        ),
        Case(
            description="string literal with false condition",
            context={},
            expression="'foo' if false",
            expect=Undefined(""),
        ),
        Case(
            description="string literal with false condition and alternative",
            context={},
            expression="'foo' if false else 'bar'",
            expect="bar",
        ),
        Case(
            description="object and condition from context",
            context={"settings": {"foo": True}, "greeting": "hello"},
            expression="greeting if settings.foo else 'bar'",
            expect="hello",
        ),
        Case(
            description="object and condition from context with tail filter",
            context={"settings": {"foo": True}, "greeting": "hello"},
            expression="greeting if settings.foo else 'bar' || upcase",
            expect="HELLO",
        ),
        Case(
            description="object filter with true condition",
            context={},
            expression="'foo' | upcase if true else 'bar'",
            expect="FOO",
        ),
        Case(
            description="object filter with false condition",
            context={},
            expression="'foo' | upcase if false else 'bar'",
            expect="bar",
        ),
    ]

    def test_eval_conditional(self) -> None:
        """Test that we can evaluate conditional expressions."""
        env = Environment()

        for case in self.test_cases:
            with self.subTest(msg=case.description):
                context = Context(env, globals=case.context)
                expr = parse_conditional_expression(case.expression)
                self.assertEqual(expr.evaluate(context), case.expect)

    def test_eval_with_parens(self) -> None:
        """Test that we can evaluate conditional expressions that support logical
        `not` and grouping terms with parentheses."""
        env = Environment()

        for case in self.test_cases:
            with self.subTest(msg=case.description):
                context = Context(env, globals=case.context)
                expr = parse_conditional_expression_with_parens(case.expression)
                self.assertEqual(expr.evaluate(context), case.expect)


class EvalConditionalNotExpressionTestCase(unittest.TestCase):
    """Test cases for evaluating non-standard conditional expressions, including
    logical `not` and grouping terms with parentheses."""

    test_cases = [
        Case(
            description="string literal with true condition",
            context={},
            expression="'foo' if not true",
            expect=Undefined(""),
        ),
        Case(
            description="string literal with false condition",
            context={},
            expression="'foo' if not false",
            expect="foo",
        ),
        Case(
            description="string literal with false condition and alternative",
            context={},
            expression="'foo' if not false else 'bar'",
            expect="foo",
        ),
        Case(
            description="object and condition from context",
            context={"settings": {"foo": True}, "greeting": "hello"},
            expression="greeting if not settings.foo else 'bar'",
            expect="bar",
        ),
        Case(
            description="object and condition from context with tail filter",
            context={"settings": {"foo": True}, "greeting": "hello"},
            expression="greeting if not settings.foo else 'bar' || upcase",
            expect="BAR",
        ),
        Case(
            description="object filter with true condition",
            context={},
            expression="'foo' | upcase if not true else 'bar'",
            expect="bar",
        ),
        Case(
            description="object filter with false condition",
            context={},
            expression="'foo' | upcase if not false else 'bar'",
            expect="FOO",
        ),
    ]

    def test_eval_with_parens(self) -> None:
        """Test that we can evaluate conditional expressions that support logical
        `not` and grouping terms with parentheses."""
        env = Environment()

        for case in self.test_cases:
            with self.subTest(msg=case.description):
                context = Context(env, globals=case.context)
                expr = parse_conditional_expression_with_parens(case.expression)
                self.assertEqual(expr.evaluate(context), case.expect)


class MalformedConditionalExpressionTestCase(unittest.TestCase):
    """Test cases for malformed conditional expressions."""

    def test_missing_condition(self) -> None:
        """Test that we handle conditional expressions with missing conditions."""
        env = Environment(undefined=Undefined)
        context = Context(env)

        # Condition defaults to `Undefined`
        expr = parse_conditional_expression("'foo' if")
        self.assertEqual(expr.evaluate(context), Undefined(""))

        # Same for extra boolean expressions
        expr = parse_conditional_expression_with_parens("'foo' if")
        self.assertEqual(expr.evaluate(context), Undefined(""))

    def test_missing_alternative(self) -> None:
        """Test that we handle conditional expressions with a missing alternative."""
        env = Environment(undefined=Undefined)
        context = Context(env)

        # Alternative defaults to `Undefined`
        expr = parse_conditional_expression("'foo' if false else")
        self.assertEqual(expr.evaluate(context), Undefined(""))

        # Same for extra boolean expressions
        expr = parse_conditional_expression_with_parens("'foo' if false else")
        self.assertEqual(expr.evaluate(context), Undefined(""))

    def test_missing_condition_followed_by_else(self) -> None:
        """Test that we handle missing boolean expressions."""
        env = Environment(undefined=Undefined)
        context = Context(env)

        # Alternative defaults to `Undefined`
        expr = parse_conditional_expression("'foo' if else 'bar'")
        self.assertEqual(expr.evaluate(context), "bar")

        # Same for extra boolean expressions
        expr = parse_conditional_expression_with_parens("'foo' if else 'bar'")
        self.assertEqual(expr.evaluate(context), "bar")



"""Tag and node definition for the built-in "echo" tag."""

import sys


from liquid.ast import Node
from liquid.builtin.statement import StatementNode

from liquid.expression import NIL
from liquid.expression import Expression
from liquid.parse import expect

from liquid.stream import TokenStream
from liquid.tag import Tag

from liquid.token import TOKEN_TAG
from liquid.token import TOKEN_EXPRESSION
from liquid.token import TOKEN_EOF

TAG_ECHO = sys.intern("echo")


class EchoNode(StatementNode):
    """Parse tree node for the built-in "echo" tag."""

    def __repr__(self) -> str:  # pragma: no cover
        return f"EchoNode(tok={self.tok}, expression={self.expression!r})"


class EchoTag(Tag):
    """The built-in "echo" tag."""

    name = TAG_ECHO
    block = False

    def _parse_expression(self, value: str) -> Expression:
        return self.env.parse_filtered_expression_value(value)

    def parse(self, stream: TokenStream) -> Node:
        expect(stream, TOKEN_TAG, value=TAG_ECHO)
        tok = stream.current
        stream.next_token()

        if stream.current.type == TOKEN_EOF:
            expr: Expression = NIL
        else:
            expect(stream, TOKEN_EXPRESSION)
            expr = self._parse_expression(stream.current.value)
        return EchoNode(tok, expression=expr)



"""Token definitions"""

import sys
from typing import NamedTuple

__all__ = (
    "TOKEN_ILLEGAL",
    "TOKEN_INITIAL",
    "TOKEN_EOF",
    "TOKEN_NEWLINE",
    "TOKEN_SKIP",
    "TOKEN_TAG",
    "TOKEN_EXPRESSION",
    "TOKEN_STATEMENT",
    "TOKEN_LITERAL",
    "TOKEN_IDENTIFIER",
    "TOKEN_STRING",
    "TOKEN_INTEGER",
    "TOKEN_FLOAT",
    "TOKEN_EMPTY",
    "TOKEN_NIL",
    "TOKEN_NULL",
    "TOKEN_BLANK",
    "TOKEN_WITH",
    "TOKEN_FOR",
    "TOKEN_AS",
    "TOKEN_BY",
    "TOKEN_NEGATIVE",
    "TOKEN_TRUE",
    "TOKEN_FALSE",
    "TOKEN_CONTAINS",
    "TOKEN_IN",
    "TOKEN_LPAREN",
    "TOKEN_RPAREN",
    "TOKEN_RANGE",
    "TOKEN_RANGE_LITERAL",
    "TOKEN_LIMIT",
    "TOKEN_OFFSET",
    "TOKEN_REVERSED",
    "TOKEN_CONTINUE",
    "TOKEN_COLS",
    "TOKEN_PIPE",
    "TOKEN_COLON",
    "TOKEN_COMMA",
    "TOKEN_DOT",
    "TOKEN_LBRACKET",
    "TOKEN_RBRACKET",
    "TOKEN_ASSIGN",
    "TOKEN_AND",
    "TOKEN_OR",
    "TOKEN_NOT",
    "TOKEN_EQ",
    "TOKEN_NE",
    "TOKEN_LG",
    "TOKEN_LT",
    "TOKEN_GT",
    "TOKEN_LE",
    "TOKEN_GE",
    "operators",
    "reverse_operators",
    "Token",
)

TOKEN_ILLEGAL = sys.intern("illegal")
TOKEN_INITIAL = sys.intern("initial")
TOKEN_EOF = sys.intern("eof")
TOKEN_NEWLINE = sys.intern("newline")
TOKEN_SKIP = sys.intern("skip")

TOKEN_TAG = sys.intern("tag")
TOKEN_EXPRESSION = sys.intern("expression")
TOKEN_STATEMENT = sys.intern("statement")
TOKEN_LITERAL = sys.intern("literal")

TOKEN_IDENTIFIER = sys.intern("identifier")
TOKEN_IDENTINDEX = sys.intern("identindex")
TOKEN_IDENTSTRING = sys.intern("identstring")
TOKEN_STRING = sys.intern("string")
TOKEN_INTEGER = sys.intern("integer")
TOKEN_FLOAT = sys.intern("float")
TOKEN_EMPTY = sys.intern("empty")
TOKEN_NIL = sys.intern("nil")
TOKEN_NULL = sys.intern("null")
TOKEN_BLANK = sys.intern("blank")

# "include" and "render" keywords
TOKEN_WITH = sys.intern("with")
TOKEN_FOR = sys.intern("for")
TOKEN_AS = sys.intern("as")

# "paginate" keywords
TOKEN_BY = sys.intern("by")

TOKEN_NEGATIVE = sys.intern("negative")

TOKEN_TRUE = sys.intern("true")
TOKEN_FALSE = sys.intern("false")

# String, array and hash membership
TOKEN_CONTAINS = sys.intern("contains")

# Looping symbols and keywords. Use by `for` and `tablerow` tags.
TOKEN_IN = sys.intern("in")
TOKEN_LPAREN = sys.intern("lparen")
TOKEN_RPAREN = sys.intern("rparen")
TOKEN_RANGE = sys.intern("range")
TOKEN_RANGE_LITERAL = sys.intern("rangeliteral")
TOKEN_LIMIT = sys.intern("limit")
TOKEN_OFFSET = sys.intern("offset")
TOKEN_REVERSED = sys.intern("reversed")
TOKEN_CONTINUE = sys.intern("continue")

# Tablerow specific argument
TOKEN_COLS = sys.intern("cols")

# Comparison symbols and logic operators for `if` and `unless` tags.
TOKEN_EQ = sys.intern("eq")
TOKEN_NE = sys.intern("ne")
TOKEN_LG = sys.intern("ltgt")
TOKEN_LT = sys.intern("lt")
TOKEN_GT = sys.intern("gt")
TOKEN_LE = sys.intern("le")
TOKEN_GE = sys.intern("ge")
TOKEN_AND = sys.intern("and")
TOKEN_OR = sys.intern("or")
TOKEN_NOT = sys.intern("not")
TOKEN_ELSE = sys.intern("else")
TOKEN_IF = sys.intern("if")


# Filter symbols
TOKEN_PIPE = sys.intern("pipe")
TOKEN_COLON = sys.intern("colon")
TOKEN_COMMA = sys.intern("comma")
TOKEN_DPIPE = sys.intern("dpipe")

# Identifier symbols
TOKEN_DOT = sys.intern("dot")
TOKEN_LBRACKET = sys.intern("lbracket")
TOKEN_RBRACKET = sys.intern("rbracket")

# Assignment
TOKEN_ASSIGN = sys.intern("assign")

# Non-standard key/value argument separator. Standard liquid uses TOKEN_COLON.
TOKEN_EQUALS = sys.intern("equals")

operators = {
    "==": TOKEN_EQ,
    "!=": TOKEN_NE,
    "<>": TOKEN_LG,
    "<": TOKEN_LT,
    ">": TOKEN_GT,
    "<=": TOKEN_LE,
    ">=": TOKEN_GE,
    "|": TOKEN_PIPE,
    ":": TOKEN_COLON,
    ",": TOKEN_COMMA,
    ".": TOKEN_DOT,
    "-": TOKEN_NEGATIVE,
    "(": TOKEN_LPAREN,
    ")": TOKEN_RPAREN,
    "..": TOKEN_RANGE,
    "[": TOKEN_LBRACKET,
    "]": TOKEN_RBRACKET,
    "=": TOKEN_ASSIGN,
}

reverse_operators = {v: k for k, v in operators.items()}


class Token(NamedTuple):
    """A token, as generated by the template or expression lexers."""

    linenum: int
    type: str
    value: str

    def test(self, typ: str) -> bool:
        """Return `True` if this token's type matches the `typ` argument."""
        return self.type == typ

    def istag(self, name: str) -> bool:
        """Return `True` if this token is a tag and has the given name."""
        return self.type == TOKEN_TAG and self.value == name



"""Bad context test cases."""

from unittest import TestCase

from typing import NamedTuple
from typing import Type

from liquid.context import builtin
from liquid.context import get_item
from liquid.context import _undefined
from liquid.context import ReadOnlyChainMap
from liquid.environment import Environment

from liquid.exceptions import LiquidTypeError
from liquid.exceptions import lookup_warning

from liquid.mode import Mode


class Case(NamedTuple):
    """Table driven test case helper."""

    description: str
    template: str
    expect_exception: Type[Exception]
    expect_msg: str
    expect_render: str = ""


class BadContextTemplateTestCase(TestCase):
    """Bad context test case."""

    def _test(self, test_cases, mode=Mode.STRICT):
        """Helper method for running lists of `Case`s"""
        env = Environment()
        env.mode = mode

        global_context = {"arr": [], "hash": {}}

        for case in test_cases:
            with self.subTest(msg=case.description):
                if mode == Mode.STRICT:
                    with self.assertRaises(case.expect_exception) as raised:
                        template = env.from_string(
                            case.template, globals=global_context
                        )
                        result = template.render()

                    self.assertEqual(str(raised.exception), case.expect_msg)

                elif mode == Mode.WARN:
                    with self.assertWarns(lookup_warning(case.expect_exception)):
                        template = env.from_string(
                            case.template, globals=global_context
                        )
                        result = template.render()

                elif mode == Mode.LAX:
                    template = env.from_string(case.template, globals=global_context)
                    result = template.render()
                    self.assertEqual(result, case.expect_render)

    def test_bad_context(self):
        """Test that we handle render time errors due to incorrect context."""
        test_cases = [
            Case(
                description="array less than hash",
                template="{% if arr < hash %}foo{% endif %}",
                expect_exception=LiquidTypeError,
                expect_msg=r"invalid operator for types '[] < {}', on line 1",
            ),
        ]

        self._test(test_cases, mode=Mode.STRICT)
        self._test(test_cases, mode=Mode.WARN)
        self._test(test_cases, mode=Mode.LAX)


class ReadOnlyChainMapTestCase(TestCase):
    """Read only chain map test case."""

    def test_get(self):
        """Test that we can get items from a chain map."""
        test_cases = [
            {
                "description": "earlier maps take priority",
                "maps": ({"foo": 1}, {"foo": 2}),
                "expect": 1,
            },
            {
                "description": "fall back top later maps",
                "maps": ({"bar": 1}, {"foo": 2}),
                "expect": 2,
            },
            {
                "description": "default to None",
                "maps": ({"bar": 1}, {"bar": 2}),
                "expect": None,
            },
        ]

        for case in test_cases:
            with self.subTest(msg=case["description"]):
                chain_map = ReadOnlyChainMap(*case["maps"])
                self.assertEqual(chain_map.get("foo"), case["expect"])

    def test_iter(self):
        """Test that we can iterate a chain map."""
        chain_map = ReadOnlyChainMap({"foo": 1}, {"bar": 2}, {"foo": 3})
        self.assertEqual(list(chain_map), ["foo", "bar", "foo"])


class ChainedItemGetterTestCase(TestCase):
    """Chained item getter test case."""

    def test_get_item(self):
        """Test that we can get nested items."""
        test_cases = [
            {
                "description": "single string key",
                "obj": {"foo": 1},
                "key": ["foo"],
                "expect": 1,
            },
            {
                "description": "chained string key",
                "obj": {"foo": {"bar": 2}},
                "key": ["foo", "bar"],
                "expect": 2,
            },
            {
                "description": "single int key",
                "obj": ["foo", "bar"],
                "key": [0],
                "expect": "foo",
            },
            {
                "description": "chained string and int key",
                "obj": {"foo": [1, 2]},
                "key": ["foo", 1],
                "expect": 2,
            },
            {
                "description": "default to undefined",
                "obj": {"foo": 1},
                "key": ["no", "such", "thing"],
                "expect": _undefined,
            },
        ]

        for case in test_cases:
            with self.subTest(msg=case["description"]):
                self.assertEqual(get_item(case["obj"], *case["key"]), case["expect"])


class BuiltinDynamicScopeTestCase(TestCase):
    """Built-in dynamic scope test case."""

    def test_builtin_contains_now(self):
        """Test that `now` is in the builtin scope."""
        self.assertTrue("now" in builtin)

    def test_builtin_contains_today(self):
        """Test that `today` is in the builtin scope."""
        self.assertTrue("today" in builtin)

    def test_builtin_not_contains(self):
        """Test that garbage is not in the builtin scope."""
        self.assertFalse("foo" in builtin)

    def test_builtin_length(self):
        """Test that builtin has a length."""
        self.assertEqual(len(builtin), 2)

    def test_builtin_iter(self):
        """Test that builtin has a length."""
        self.assertEqual(list(builtin), ["now", "today"])



"""Tag and node definition for the built-in "cycle" tag."""
import sys

from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from typing import TextIO

from liquid.ast import ChildNode
from liquid.ast import Node

from liquid.context import Context
from liquid.exceptions import LiquidSyntaxError
from liquid.expression import Expression

from liquid.parse import expect
from liquid.stream import TokenStream
from liquid.tag import Tag

from liquid.token import Token
from liquid.token import TOKEN_EXPRESSION
from liquid.token import TOKEN_EOF
from liquid.token import TOKEN_COMMA
from liquid.token import TOKEN_COLON

from liquid.expressions import Token as ExprToken
from liquid.expressions import TokenStream as ExprTokenStream
from liquid.expressions.common import parse_string_or_identifier
from liquid.expressions.filtered.lex import tokenize
from liquid.expressions.filtered.parse import parse_obj

TAG_CYCLE = sys.intern("cycle")


class CycleNode(Node):
    """Parse tree node for the built-in "cycle" tag."""

    __slots__ = ("tok", "group", "args", "key")

    def __init__(self, tok: Token, group: Optional[Expression], args: List[Expression]):
        self.tok = tok
        self.group = group
        self.args = args

    def __str__(self) -> str:
        buf = ["cycle ("]
        if self.group:
            buf.append(f"{self.group}: ")

        buf.append(", ".join([str(arg) for arg in self.args]))
        buf.append(")")
        return "".join(buf)

    def render_to_output(self, context: Context, buffer: TextIO) -> Optional[bool]:
        if self.group:
            group_name = str(self.group.evaluate(context))
        else:
            group_name = ""

        args = [arg.evaluate(context) for arg in self.args]
        buffer.write(str(next(context.cycle(group_name, args))))
        return True

    async def render_to_output_async(
        self, context: Context, buffer: TextIO
    ) -> Optional[bool]:
        if self.group:
            group_name = str(await self.group.evaluate_async(context))
        else:
            group_name = ""

        args = [await arg.evaluate_async(context) for arg in self.args]
        buffer.write(str(next(context.cycle(group_name, args))))
        return True

    def children(self) -> List[ChildNode]:
        _children: List[ChildNode] = []
        if self.group:
            _children.append(
                ChildNode(
                    linenum=self.tok.linenum,
                    node=None,
                    expression=self.group,
                )
            )
        for arg in self.args:
            _children.append(ChildNode(linenum=self.tok.linenum, expression=arg))
        return _children


def split_at_first_colon(tokens: Iterable[ExprToken]) -> Iterator[List[ExprToken]]:
    """Split tokens on into lists, using TOKEN_COLON as the delimiter."""
    buf: List[ExprToken] = []
    for token in tokens:
        if token[1] == TOKEN_COLON:
            yield buf
            buf = []
        else:
            buf.append(token)
    yield buf


class CycleTag(Tag):
    """The built-in "cycle" tag."""

    name = TAG_CYCLE
    block = False

    def parse(self, stream: TokenStream) -> CycleNode:
        tok = next(stream)
        expect(stream, TOKEN_EXPRESSION)
        tokens = tokenize(stream.current.value, linenum=tok.linenum)
        group_name: Optional[Expression] = None

        parts = list(split_at_first_colon(tokens))
        if len(parts) == 2:
            group_name = parse_string_or_identifier(ExprTokenStream(iter(parts[0])))

        args: List[Expression] = []
        expr_stream = ExprTokenStream(iter(parts[-1]))
        while True:
            args.append(parse_obj(expr_stream))
            next(expr_stream)

            typ = expr_stream.current[1]
            if typ == TOKEN_COMMA:
                next(expr_stream)
            elif typ == TOKEN_EOF:
                break
            else:
                raise LiquidSyntaxError(
                    f"expected a comma separated list of arguments, found {typ}",
                    linenum=expr_stream.current[0],
                )

        return CycleNode(tok, group_name, args)



"""Miscellaneous filters."""
from __future__ import annotations

import datetime
import decimal
import functools

from typing import Any
from typing import Union
from typing import TYPE_CHECKING

from dateutil import parser

try:
    from markupsafe import Markup
except ImportError:
    from liquid.exceptions import Markup  # type: ignore

from liquid import is_undefined

from liquid.filter import liquid_filter
from liquid.filter import with_environment

from liquid.exceptions import FilterArgumentError
from liquid.expression import EMPTY

if TYPE_CHECKING:
    from liquid import Environment


@liquid_filter
def size(obj: Any) -> int:
    """Return the length of an array or string."""
    try:
        return len(obj)
    except TypeError:
        return 0


@liquid_filter
def default(obj: Any, default_: object = "", *, allow_false: bool = False) -> Any:
    """Return a default value if the input is nil, false, or empty."""
    _obj = obj

    # Return the default value immediately if the object defines a
    # `force_liquid_default` property.
    if hasattr(obj, "force_liquid_default") and obj.force_liquid_default:
        return default_

    if hasattr(obj, "__liquid__"):
        _obj = obj.__liquid__()

    # Liquid 0, 0.0, 0b0, 0X0, 0o0 and Decimal("0") are not falsy.
    if not isinstance(obj, bool) and isinstance(obj, (int, float, decimal.Decimal)):
        return obj

    if allow_false is True and _obj is False:
        return obj

    if _obj in (None, False, EMPTY):
        return default_

    return obj


@with_environment
@liquid_filter
@functools.lru_cache(maxsize=10)
def date(
    dat: Union[datetime.datetime, str, int],
    fmt: str,
    *,
    environment: Environment,
) -> str:
    """Formats a datetime according the the given format string."""
    if is_undefined(dat):
        return ""

    if is_undefined(fmt):
        return str(dat)

    if isinstance(dat, str):
        if dat in ("now", "today"):
            dat = datetime.datetime.now()
        elif dat.isdigit():
            # The reference implementation does not support string
            # representations of negative integers either.
            dat = datetime.datetime.fromtimestamp(int(dat))
        else:
            try:
                dat = parser.parse(dat)
            except parser.ParserError:
                # Input is returned unchanged. This is consistent
                # with the reference implementation.
                return str(dat)
    elif isinstance(dat, int):
        try:
            dat = datetime.datetime.fromtimestamp(dat)
        except (OverflowError, OSError):
            # Testing on Windows shows that it can't handle some
            # negative integers.
            return str(dat)

    if not isinstance(dat, (datetime.datetime, datetime.date)):
        raise FilterArgumentError(
            f"date expected datetime.datetime, found {type(dat).__name__}"
        )

    rv = dat.strftime(fmt)

    if environment.autoescape and isinstance(fmt, Markup):
        return Markup(rv)
    return rv



"""Golden tests cases for testing liquid's built-in `base64_decode` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="from string",
        template=r'{{ "XyMvLg==" | base64_decode }}',
        expect="_#/.",
    ),
    Case(
        description="from string with URL unsafe",
        template=r"{{ a | base64_decode }}",
        expect=(
            r"abcdefghijklmnopqrstuvwxyz "
            r"ABCDEFGHIJKLMNOPQRSTUVWXYZ "
            r"1234567890 !@#$%^&*()-=_+/?.:;[]{}\|"
        ),
        globals={
            "a": (
                "YWJjZGVmZ2hpamtsbW5vcHFyc3R1dnd4eXogQUJDREVGR0hJSktMTU5PUFFSU1RVV"
                "ldYWVogMTIzNDU2Nzg5MCAhQCMkJV4mKigpLT1fKy8/Ljo7W117fVx8"
            )
        },
    ),
    Case(
        description="not a string",
        template=r"{{ 5 | base64_decode }}",
        expect="",
        error=True,
    ),
    Case(
        description="unexpected argument",
        template=r'{{ "hello" | base64_decode: 5 }}',
        expect="",
        error=True,
    ),
    Case(
        description="undefined left value",
        template=r"{{ nosuchthing | base64_decode }}",
        expect="",
    ),
]



"""Legacy iterable drop base class."""
from typing import Any
from typing import Collection
from abc import ABC
from abc import abstractmethod


# pylint: disable=too-few-public-methods
class IterableDrop(ABC, Collection[Any]):
    """Inherit from this class if you want the drop to be iterable from
    tags like 'include' and 'render'."""

    @abstractmethod
    def step(self, item: Any) -> None:
        """Step the iterator."""



"""Test cases for analyzing a template given some render context data."""

import asyncio
import unittest

from typing import Dict

from liquid import Context
from liquid import DictLoader
from liquid import Environment
from liquid import Template

from liquid.filter import with_context
from liquid.template import BoundTemplate


class ContextualAnalysisTestCase(unittest.TestCase):
    """Contextual template analysis test cases."""

    def _test(
        self,
        template: BoundTemplate,
        data: Dict[str, object],
        variables: Dict[str, int],
        assigns: Dict[str, int],
        undefined: Dict[str, int],
    ) -> None:
        refs = template.analyze_with_context(**data)
        self.assertEqual(refs.all_variables, variables)
        self.assertEqual(refs.local_variables, assigns)
        self.assertEqual(refs.undefined_variables, undefined)

        async def coro():
            return await template.analyze_with_context_async(**data)

        refs = asyncio.run(coro())
        self.assertEqual(refs.all_variables, variables)
        self.assertEqual(refs.local_variables, assigns)
        self.assertEqual(refs.undefined_variables, undefined)

    def test_analyze_output(self):
        """Test that we count references to variables in output statements."""
        template = Template("{{ x | default: y, allow_false: z }}")
        data = {"y": 1, "z": 2}
        expect_variables = {"x": 1, "y": 1, "z": 1}
        expect_locals = {}
        expect_undefined = {"x": 1}

        self._test(template, data, expect_variables, expect_locals, expect_undefined)

    def test_visit_branches(self):
        """Test that we only count references to variables in visited branches."""
        template = Template(
            "{% if false %}{{ x | default: y, allow_false: z }}{% endif %}"
        )
        data = {"y": 1, "z": 2}
        expect_variables = {}
        expect_locals = {}
        expect_undefined = {}

        self._test(template, data, expect_variables, expect_locals, expect_undefined)

    def test_analyze_assign(self):
        """Test that we count references to variables in assignment tags."""
        template = Template("{% assign x = a %}{{ x | default: y, allow_false: z }}")
        data = {"a": 1, "y": 1, "z": 2}
        expect_variables = {"x": 1, "y": 1, "z": 1, "a": 1}
        expect_locals = {"x": 1}
        expect_undefined = {}

        self._test(template, data, expect_variables, expect_locals, expect_undefined)

    def test_count_number_of_references(self):
        """Test that we count multiple references to the same variable name."""
        template = Template("{{ x | default: y, allow_false: z }}{{ x | append: x }}")
        data = {"y": 1, "z": 2}
        expect_variables = {"x": 3, "y": 1, "z": 1}
        expect_locals = {}
        expect_undefined = {"x": 3}

        self._test(template, data, expect_variables, expect_locals, expect_undefined)

    def test_analyze_included_template(self):
        """Test that we count references to variables in included templates."""
        loader = DictLoader({"foo": "{{ x | append: y }}"})
        env = Environment(loader=loader)

        template = env.from_string("{{ x }}{% include 'foo' %}")
        data = {"y": 1, "z": 2}
        expect_variables = {"x": 2, "y": 1}
        expect_locals = {}
        expect_undefined = {"x": 2}

        self._test(template, data, expect_variables, expect_locals, expect_undefined)

    def test_analyze_rendered_template(self):
        """Test that we count references to variables in rendered templates."""
        loader = DictLoader({"foo": "{{ x | append: y }}"})
        env = Environment(loader=loader)

        template = env.from_string("{{ x }}{% render 'foo' %}")
        data = {"y": 1, "z": 2}
        expect_variables = {"x": 2, "y": 1}
        expect_locals = {}
        expect_undefined = {"x": 2}

        self._test(template, data, expect_variables, expect_locals, expect_undefined)

    def test_analyze_loop(self):
        """Test that we count variables in every iteration of a for loop."""
        template = Template(
            "{% for x in (1..5) %}{{ forloop.index }} {{ x }}{% endfor %}"
        )
        data = {"y": 1, "z": 2}
        expect_variables = {"x": 5, "forloop.index": 5}
        expect_locals = {}
        expect_undefined = {}

        self._test(template, data, expect_variables, expect_locals, expect_undefined)

    def test_analyze_broken_loop(self):
        """Test that we don't count variables after a loop break."""
        template = Template(
            "{% for x in (1..5) %}"
            "{{ forloop.index }} {{ x }}"
            "{% if forloop.index == 3 %}{% break %}{% endif %}"
            "{% endfor %}"
        )
        data = {"y": 1, "z": 2}
        expect_variables = {"x": 3, "forloop.index": 6}
        expect_locals = {}
        expect_undefined = {}

        self._test(template, data, expect_variables, expect_locals, expect_undefined)

    def test_analyze_indices(self):
        """Test that we count variables that include indices."""
        template = Template("{{ x.y[0].foo }}")
        data = {"x": {"y": [{"foo": 5}]}}
        expect_variables = {"x.y[0].foo": 1}
        expect_locals = {}
        expect_undefined = {}

        self._test(template, data, expect_variables, expect_locals, expect_undefined)

    def test_analyze_nested_identifiers(self):
        """Test that we count variables that nest identifiers."""
        template = Template("{{ x[y].foo }}")
        data = {"x": [{"foo": 5}], "y": 0}
        expect_variables = {"x[0].foo": 1, "y": 1}
        expect_locals = {}
        expect_undefined = {}

        self._test(template, data, expect_variables, expect_locals, expect_undefined)

    def test_analyze_quoted_identifiers(self):
        """Test that we count variables that quote identifiers."""
        template = Template("{{ x['y z'].foo }}")
        data = {"x": {"y z": {"foo": 5}}}
        expect_variables = {'x["y z"].foo': 1}
        expect_locals = {}
        expect_undefined = {}

        self._test(template, data, expect_variables, expect_locals, expect_undefined)

    def test_count_increment_tags(self):
        """Test that we count variable names from increment tags."""
        template = Template("{% increment x %}")
        data = {}
        expect_variables = {}
        expect_locals = {"x": 1}
        expect_undefined = {}

        self._test(template, data, expect_variables, expect_locals, expect_undefined)

    def test_count_decrement_tags(self):
        """Test that we count variable names from decrement tags."""
        template = Template("{% decrement x %}")
        data = {}
        expect_variables = {}
        expect_locals = {"x": 1}
        expect_undefined = {}

        self._test(template, data, expect_variables, expect_locals, expect_undefined)

    def test_explicit_resolve(self):
        """Test that calls to `Context.resolve()` are counted."""

        @with_context
        def mock_filter(val, *, context: Context) -> str:
            context.resolve("x")
            return f"{val} mock"

        env = Environment()
        env.add_filter("mock", mock_filter)
        template = env.from_string("{{ 'foo' | mock }}")

        data = {"x": "foo"}
        expect_variables = {"x": 1}
        expect_locals = {}
        expect_undefined = {}

        self._test(template, data, expect_variables, expect_locals, expect_undefined)
